function animateSimulation(t, graphdata, indices, matrixdata, n_rows, n_cols, opt_struct, cont_from)
	if isempty(indices), indices = (1:size(graphdata,2))'; end
	if isnan(indices), indices = []; end
	
	if iscell(indices)
		cc = length(indices); 
	else
		cc = size(indices, 1);
	end
	
	if nargin < 7, opt_struct = []; end
		if ~isfield(opt_struct, 'skip'), opt_struct.skip = 25; end
		if ~isfield(opt_struct, 'movie_fps'), opt_struct.movie_fps = 25; end
		if ~isfield(opt_struct, 'alt_layout'), opt_struct.alt_layout = 0; end
		if ~isfield(opt_struct, 'grid_spacing'), opt_struct.grid_spacing = 1; end
		if ~isfield(opt_struct, 'trace_size'), opt_struct.trace_size = ones(cc,1); end
		if ~isfield(opt_struct, 'font_size'), opt_struct.font_size = 10; end

	if nargin < 8, cont_from = 1; end
	
	figure;
		set(gcf, 'Position', [200, 200, 700, 700]);
		if isempty(indices), set(gcf, 'Position', [200,200,300,700]); end
		
	jointpos = [0, 1E6, 0, 1E6];
	axeslims = zeros(cc,4);
	for ii = 1:cc
		if opt_struct.alt_layout == 1
			subplot(3*cc,1,ii);
		else
			%subplot(cc,3,[3*(ii-1)+1, 3*(ii-1)+2]);
			if ii == 1
				currowstart = 1;
			else
				currowstart = 1 + sum(opt_struct.trace_size(1:ii-1));
			end
			currowend = sum(opt_struct.trace_size(1:ii));
			indxs = sort([(3*(currowstart-1)+1):3:(3*(currowend-1)+1), 1+((3*(currowstart-1)+1):3:(3*(currowend-1)+1))])
			subplot(sum(opt_struct.trace_size),3,indxs);
		end
		
		if iscell(indices)
			if isfield(opt_struct, 'linecolor')
				if length(opt_struct.linecolor) > 1
					cax = plot(t, graphdata(:, indices{ii}), 'LineWidth', 2, 'Color', opt_struct.linecolor(ii));
				else
					cax = plot(t, graphdata(:, indices{ii}), 'LineWidth', 2, 'Color', opt_struct.linecolor);
				end
			else
				cax = plot(t, graphdata(:, indices{ii}), 'LineWidth', 2);
			end
		else
			if isfield(opt_struct, 'linecolor')
				if length(opt_struct.linecolor) > 1
					cax = plot(t, graphdata(:, indices(ii,:)), 'LineWidth', 2, 'Color', opt_struct.linecolor(ii));
				else
					cax = plot(t, graphdata(:, indices(ii,:)), 'LineWidth', 2, 'Color', opt_struct.linecolor);
				end
			else
				cax = plot(t, graphdata(:, indices(ii,:)), 'LineWidth', 2);		
			end
		end
		
		set(gca, 'FontSize', opt_struct.font_size);
		
		if isfield(opt_struct, 'ylim')
			if size(opt_struct.ylim,1) > 1
				ylim([opt_struct.ylim(ii,1), opt_struct.ylim(ii,2)]);
			else
				ylim([opt_struct.ylim(1,1), opt_struct.ylim(1,2)]);
			end
		end
		
		limits = axis();
		axeslims(ii,:) = limits;
		%pos = get(gca, 'Position')
		%if pos(1) > jointpos(1), jointpos(1) = pos(1); end
		%if pos(2) < jointpos(2), jointpos(2) = pos(2); end
		%if pos(3) > jointpos(3), jointpos(3) = pos(3); end
		%if pos(4) < jointpos(4), jointpos(4) = pos(4); end
				
		list_of_cax{ii} = cax;
		list_of_gca{ii} = gca;
		
		if ii == cc && isfield(opt_struct, 'xlabel')
			xlabel(strtrim(char(opt_struct.xlabel)), 'FontSize', opt_struct.font_size+2, 'FontWeight', 'bold'); 
		end
		if isfield(opt_struct, 'ylabel')
			ylabel(strtrim(char(opt_struct.ylabel(ii,:))), 'FontSize', opt_struct.font_size+2, 'FontWeight', 'bold'); 
			if isfield(opt_struct, 'ylabel_pos')
				labh = get(gca,'YLabel');
				crpos = get(labh,'Position')
				%set(labh,'Position',get(labh,'Position') - [30, 0, 0])
				text(opt_struct.ylabel_pos, crpos(2), strtrim(char(opt_struct.ylabel(ii,:))), 'FontSize', opt_struct.font_size+2, 'FontWeight', 'bold', 'HorizontalAlignment', 'center', 'Rotation', 90);
				ylabel('     ');
			end
		end
		
		if isfield(opt_struct, 'plotfunc')
			feval(opt_struct.plotfunc, ii, gca, cax);
		end
	end
	
	if ~isempty(indices)
		if opt_struct.alt_layout == 1
			subplot(3*cc, 1, (cc+1):1:(3*cc));
		else
			subplot(cc, 3, 3:3:(cc*3));
		end
	end
	
		mxv = max(max(matrixdata)), if(isfield(opt_struct, 'max')) mxv = opt_struct.max; end
		miv = min(min(matrixdata)), if(isfield(opt_struct, 'min')) miv = opt_struct.min; end
		
		ih = imagesc(reshape(matrixdata(1,:), n_cols, n_rows)', [miv mxv]);
			colorbar
			grid on;
			if opt_struct.alt_layout == 1
				th = xlabel(sprintf('t = %0.2f', t(1)), 'FontSize', opt_struct.font_size);
				set(gca, 'ytick', 0.5:opt_struct.grid_spacing:(n_rows+0.5));
				set(gca, 'xtick', 0.5:opt_struct.grid_spacing:(n_cols+0.5));
			else
				th = title(sprintf('t = %0.2f', t(1)), 'FontSize', opt_struct.font_size);
				set(gca, 'ytick', 0.5:1:(n_rows+0.5));
				set(gca, 'xtick', 0.5:2:(n_cols+0.5));
			end	
			
			set(gca, 'xticklabel', '');
			set(gca, 'yticklabel', '');
			set(gca, 'tickdir', 'out');
			
			if isfield(opt_struct, 'show_indices') && ~isempty(opt_struct.show_indices)
				for ii = 1:length(opt_struct.show_indices)
					%convert index to 2D pos
					r_ind = floor(opt_struct.show_indices(ii) / n_cols);
					c_ind = opt_struct.show_indices(ii) - r_ind * n_cols;
					
					h = rectangle('Position', [c_ind + 1 - 0.5, r_ind + 1 - 0.5, 1.0, 1.0], 'Curvature', 1.0, 'Edgecolor', 'w');
					set(h, 'LineWidth', 3.0);
				end
			end
			
	if isfield(opt_struct, 'movie_file')
		%movie('init', 'data/Octave_Movie_Data/Temp.avi');
		%avifile('data/Octave_Movie_Data/Temp.avi');
		ppat = 'data/Octave_Movie_Data/Temp_%06d.pdf';				
		fn = sprintf(ppat, 0);
		
		if ~isempty(indices)
			print(fn, '-dpdf', '-S1500,1500');			
		else
			print(fn, '-dpdf', '-S640,1500');
		end
		curframe = cont_from - 1;	
	end
	
	for i=(1+(cont_from-1)*opt_struct.skip):opt_struct.skip:length(t)
		curr_data = reshape(matrixdata(i,:), n_cols, n_rows)';
			
		set(ih, 'cdata', curr_data);
		%if opt_struct.alt_layout == 1
		%	set(ih, 'xlabel', sprintf('t = %0.2f', t(i)));
		%else
			set(th, 'string', sprintf('t = %0.2f', t(i)));
		%end
		
		for j=1:cc
			for kk=1:length(list_of_cax{j})
				set(list_of_cax{j}(kk), 'xdata', t(1:i));
				if iscell(indices)
					set(list_of_cax{j}(kk), 'ydata', graphdata(1:i,indices{j}(kk)));
				else
					set(list_of_cax{j}(kk), 'ydata', graphdata(1:i,indices(j,kk)));
				end				
			end
			set(list_of_gca{j}, 'xlim', axeslims(j,1:2));
			set(list_of_gca{j}, 'ylim', axeslims(j,3:4));
			
		end				
		
		if isfield(opt_struct, 'movie_file')
			fn = sprintf(ppat, curframe)
			%drawnow("png", fn);			
			drawnow;
			if ~isempty(indices)
				print(fn, '-dpdf', '-S1500,1500');
				whos
			else
				print(fn, '-dpdf', '-S640,1500');
			end		
			%saveas(gcf, fn, 'png');
			curframe = curframe + 1;
		else
			drawnow;
		end
		
		if isfield(opt_struct, 'delay')
			pause(opt_struct.delay);
		end
	end
	
	%if isfield(opt_struct, 'movie_file')
	%	%movie('close', 'data/Octave_Movie_Data/Temp.avi');
	%		abs_fn = make_absolute_filename('data\Octave_Movie_Data\Temp_%6d.png');
	%		cmd = sprintf('C:\\Octave\\ffmpeg\\bin\\ffmpeg.exe -y -r %d -i %s %s', opt_struct.movie_fps, abs_fn, make_absolute_filename(opt_struct.movie_file));
	%		system(cmd);
	%		
	%		if exist(opt_struct.movie_file) == 2
	%			for i=0:curframe-1
	%				fn = sprintf(ppat, i)
	%				[errc, msg] = unlink(fn);
	%				if errc ~= 0
	%					disp(msg);
	%					pause;
	%				end
	%			end
	%		end
	%end
end