function createMeshDataFile(sourcefile, destfile, colindices, Y_spacing, APD_file, opt)
	if nargin < 6, opt = []; end
	
	%Example of colindices:
		% [2, 3, 4, 5, 6; 7, 8, 9, 10, 11] for a cell divived into 5 pieces, where cols 2,3,4,5,6 give the individual Cai and 7-11 give the individual JSR.

	if ischar(sourcefile) && exist(sourcefile) == 2
		data = dlmread(sourcefile);
	else
		data = sourcefile;
	end
	
	t = data(:,1);

	if nargin < 3 || isempty(colindices), colindices = 2:size(data,2); end
	if nargin < 4 || isempty(Y_spacing), Y_spacing = 1:size(colindices,2); end

	t_sindex = 1; t_eindex = size(data,1);
	if isfield(opt, 'xlim')
		[val, t_sindex] = min(abs(t - opt.xlim(1)));
		[val, t_eindex] = min(abs(t - opt.xlim(2)));
	end
	if isfield(opt, 'xoffset')
		t = t + opt.xoffset;
	end
	
	sheader = ['VARIABLES = "X", "Y",'];
	for i=1:size(colindices,1)
		sheader = [sheader, '"V', num2str(i), '",'];
	end

	sheader = [sheader(1:end-1), '\n']
	szone = ['ZONE I=', num2str(size(colindices,2)), ', J=', num2str(t_eindex-t_sindex+1), ', DATAPACKING=POINT\n']

	fid = fopen(destfile, 'w');
		fprintf(fid, sheader);
		fprintf(fid, szone);
	fclose(fid);

	newdata = zeros(size(data(t_sindex:t_eindex),1)*size(colindices,2),2+size(colindices,1));
		
	for j=1:size(colindices,2)
		for i=t_sindex:t_eindex
			newdata((i-t_sindex)*size(colindices,2)+j,1:2) = [t(i), Y_spacing(j)];
			for k=1:size(colindices,1)
				newdata((i-t_sindex)*size(colindices,2)+j,2+k) = data(i,colindices(k,j));
			end
		end
	end

	dlmwrite(destfile, newdata, '-append');	

	if nargin >= 5 && ~isempty(APD_file) && exist(APD_file) == 2
		data_apd = dlmread(APD_file);
		data_apd = [(1:size(data_apd,2)-1)', data_apd(2:end,2:end)'];
		
		dlmwrite([destfile(1:end-4), '_APD.txt'], data_apd);
	end
end