function waves = detectCalciumWaves(t, Ca_data, settings, options)
	if ~isfield(options, 'threshold'), options.threshold = 250; end
	if ~isfield(options, 'tdistance'), options.tdistance = 200; end
	if ~isfield(options, 'neighbordist'), options.neighbordist = 100; end

	waves = [];
	for i = 1:size(Ca_data, 2)
	
		% Find local maximum
		indices = (Ca_data(2:end-1,i) > options.threshold) & (Ca_data(2:end-1,i) >= Ca_data(1:end-2,i)) & (Ca_data(2:end-1,i) >= Ca_data(3:end,i));
		newindices = 1 + find(indices);
		
		if ~isempty(newindices)
			newindices;
			
			% Find a way to remove local maxima that are too close together, remove lowest
			indicestokeep = [];
			j = 1;
			while j <= length(newindices)
				indicestoeval = find(t(newindices(j:end)) < (t(newindices(j)) + options.tdistance)) + j - 1;
				[val, indtok] = max(Ca_data(newindices(indicestoeval),i));
				indicestokeep = [indicestokeep, newindices(indicestoeval(indtok))];
				j = j + length(indicestoeval);
			end
			newindices = indicestokeep;
			
			%Now look at neighboring units
			[x_dom,y_seg] = indexToXY(i, settings);
			%indicestokeep = [];
			%for j=1:length(newindices)
			%	if Ca_data(newindices(j),i) >= Ca_data(newindices(j),XYToIndex(x_dom-1, y_seg, settings)) && Ca_data(newindices(j),i) >= Ca_data(newindices(j),XYToIndex(x_dom+1, y_seg, settings))
			%		if Ca_data(newindices(j),i) >= Ca_data(newindices(j),XYToIndex(x_dom, y_seg-1, settings)) && Ca_data(newindices(j),i) >= Ca_data(newindices(j),XYToIndex(x_dom, y_seg+1, settings))
			%			indicestokeep = [indicestokeep, newindices(j)];
			%		end
			%	end
			%end
			%indicestokeep;
			
			if ~isempty(indicestokeep)
				waves = [waves; [t(indicestokeep), x_dom .* ones(length(indicestokeep),1), y_seg .* ones(length(indicestokeep),1), ones(length(indicestokeep),1)]];	
			end
		end
	end
	
	if ~isempty(waves)
		[vals, sortind] = sort(waves(:,1));
		waves = waves(sortind,:);
	
		for i=1:size(waves,1)
			% See if there is a point next to me that had 
			indices = find((waves(i,1) - waves(1:i-1,1)) < options.neighbordist & abs(waves(i,2) - waves(1:i-1,2)) <= 2 & abs(waves(i,3) - waves(1:i-1,3)) <= 2);
			if ~isempty(indices)
				% Is this point a source??
				if waves(indices(1),4) > 0
					% It is a source, increase its sice and make a ref to the source
					waves(indices(1),4) = waves(indices(1),4) + 1;
					waves(i,4) = -indices(1);
				else
					% It is not a source, use the reference stored to find the actual source
					waves(-waves(indices(1),4),4) = waves(-waves(indices(1),4),4) + 1;
					waves(i,4) = waves(indices(1),4);
				end
			end
		end
		
		waves;
		waves(:,1)=waves(:,1); 
    csvwrite('waves_total.txt', waves);
		waves = waves(find(waves(:,4) > 0),:);
	end

end

function [x_dom,y_seg] = indexToXY(l_index_v, settings)
	y_seg = 1 + floor((l_index_v - 0.5) / settings.N_CaDomains);
	x_dom = mod(l_index_v, settings.N_CaDomains);
	if x_dom == 0, x_dom = settings.N_CaDomains; end
end

function r_index_v = XYToIndex(x_dom, y_seg, settings)
	if x_dom < 1, x_dom = 1; end
	if x_dom > settings.N_CaDomains, x_dom = settings.N_CaDomains; end
	
	if y_seg < 1, y_seg = 1; end
	if y_seg > settings.N_Segments, y_seg = settings.N_Segments; end
	
	r_index_v = (y_seg - 1) * settings.N_CaDomains + x_dom;
end