#include "HAMM.h"
void HAMM::init(Constants *idata, Settings *isettings)
{		
	showlog = false;
	data = idata;
	settings = isettings;	

	tmaxdvdt = 0;
	valmaxdvdt = 0;
	vrest = Vm;
	vmax = 0;
	APD = 0;
	CaT_Min = 1;
	CaT_Max = -1;
	
	Ca_junc = 2.0738e-04;
	if(settings->clamp_Ca_i >= 0) Ca_junc = settings->clamp_Ca_i;
	
	double Km_SLL = data->koff_sll / data->kon_sll;
	double Km_SLH = data->koff_slh / data->kon_slh;
	
	double Buff_Ca_SLLj = data->Bmax_SLlowj * Ca_junc / (Ca_junc + Km_SLL);
	double Buff_Ca_SLHj = data->Bmax_SLhighj * Ca_junc / (Ca_junc + Km_SLH);
	double Buff_Ca_SLLsl = data->Bmax_SLlowsl * Ca_junc / (Ca_junc + Km_SLL);
	double Buff_Ca_SLHsl = data->Bmax_SLhighsl * Ca_junc / (Ca_junc + Km_SLH);

	Vm = -73.496; 
	Na_i = 9.1366; Na_srs = 9.1361; Na_sl = 9.1364; K_i = 120;
	Ca_i = 2.0738e-04; Ca_sl = 2.2918e-4;
	if(settings->clamp_Ca_i >= 0) Ca_i = settings->clamp_Ca_i;
	if(settings->clamp_Ca_i >= 0) Ca_sl = settings->clamp_Ca_i;
	Ca_sr = 0.25;
	Ca_srs = Ca_junc;
	
	Copy_Ca_sl = new double*[settings->N_Segments];
	Copy_Na_sl = new double*[settings->N_Segments];
	Copy_Ca_sr = new double*[settings->N_Segments];
	Copy_Ca_srs = new double*[settings->N_Segments];
	Copy_Ca_i = new double*[settings->N_Segments];
	Copy_Na_i = new double*[settings->N_Segments];
	Copy_Na_srs = new double*[settings->N_Segments];
					
	segments = new Segment[settings->N_Segments];
		for(int ii=0; ii<settings->N_Segments; ii++)
		{
			Copy_Ca_sl[ii] = new double[2];
			Copy_Na_sl[ii] = new double[2];
			Copy_Ca_i[ii] = new double[settings->N_CaDomains];
			Copy_Ca_sr[ii] = new double[settings->N_CaDomains];
			Copy_Ca_srs[ii] = new double[settings->N_CaDomains];
			Copy_Na_i[ii] = new double[settings->N_CaDomains];
			Copy_Na_srs[ii] = new double[settings->N_CaDomains];
			
			// Loop through the two membrane sides
			for(int jj=0; jj<2; jj++)
			{
				segments[ii].membrane[jj].m_Na = 1.8698e-02; segments[ii].membrane[jj].h_Na = 3.1846e-01; segments[ii].membrane[jj].j_Na = 2.6941e-01; 
				segments[ii].membrane[jj].mL_Na = 1.0092e-02; segments[ii].membrane[jj].hL_Na = 4.053e-02; segments[ii].membrane[jj].xr = 1.490e-3; 
				segments[ii].membrane[jj].to_a = 1.3712e-3; segments[ii].membrane[jj].to_if = 9.4622e-1; segments[ii].membrane[jj].kur_a = 3.9041e-01; 
				segments[ii].membrane[jj].kur_i = 9.5862e-01; segments[ii].membrane[jj].xs_NP = 7.4591e-3; segments[ii].membrane[jj].xs_P = 5.412034e-3;
			
				segments[ii].membrane[jj].Buff_Na_sl = 7.8794e-1;
				segments[ii].membrane[jj].Buff_Ca_SLLj = Buff_Ca_SLLj; segments[ii].membrane[jj].Buff_Ca_SLLsl = Buff_Ca_SLLsl; //2.1059e-02;
				segments[ii].membrane[jj].Buff_Ca_SLHj = Buff_Ca_SLHj; segments[ii].membrane[jj].Buff_Ca_SLHsl = Buff_Ca_SLHsl; //1.8888e-01;
				segments[ii].membrane[jj].Buff_Ca_EGTA_sl = 0;
				
				segments[ii].membrane[jj].Ca_junc = Ca_junc;
				segments[ii].membrane[jj].Ca_sl = Ca_sl;				
				segments[ii].membrane[jj].Na_sl = Na_sl;	
			}
			
			// Loop through the cytosolic Ca2+ domains
			segments[ii].caUnits = new CaDomain[settings->N_CaDomains];
			for(int jj=0; jj<settings->N_CaDomains; jj++)
			{
				segments[ii].caUnits[jj].Buff_Na_junc = 3.6109;
				segments[ii].caUnits[jj].Na_srs = Na_srs;
				segments[ii].caUnits[jj].Na_i = Na_i;				
								
				segments[ii].caUnits[jj].ICaLState_NP[0] = 1.0; segments[ii].caUnits[jj].ICaLState_NP[1] = 0.0; segments[ii].caUnits[jj].ICaLState_NP[2] = 0.0; segments[ii].caUnits[jj].ICaLState_NP[3] = 0.0;
				segments[ii].caUnits[jj].ICaLState_NP[4] = 0.0; segments[ii].caUnits[jj].ICaLState_NP[5] = 0.0; segments[ii].caUnits[jj].ICaLState_NP[6] = 0.0; segments[ii].caUnits[jj].ICaLState_NP[7] = 0.0;

				if(settings->RyRModelType == 3)
				{
					double fact = settings->numRyRs * 1.0 / (settings->N_Segments * settings->N_CaDomains);
					segments[ii].caUnits[jj].RyRState_NP[0] = 0; segments[ii].caUnits[jj].RyRState_NP[1] = 0; segments[ii].caUnits[jj].RyRState_NP[2] = 0; 
					segments[ii].caUnits[jj].RyRState_NP[3] = (int)(fact); segments[ii].caUnits[jj].RyRState_NP[4] = 0; segments[ii].caUnits[jj].RyRState_NP[5] = 0;
					segments[ii].caUnits[jj].RyRState_NP[6] = 0; segments[ii].caUnits[jj].RyRState_NP[7] = 0;
				}
				else
				{		
					segments[ii].caUnits[jj].RyRState_NP[0] = 0; segments[ii].caUnits[jj].RyRState_NP[1] = 0; segments[ii].caUnits[jj].RyRState_NP[2] = 0; 
					segments[ii].caUnits[jj].RyRState_NP[3] = 1.0; segments[ii].caUnits[jj].RyRState_NP[4] = 0; segments[ii].caUnits[jj].RyRState_NP[5] = 0;
					segments[ii].caUnits[jj].RyRState_NP[6] = 0; segments[ii].caUnits[jj].RyRState_NP[7] = 0;
				}
				segments[ii].caUnits[jj].RyR_Integr = 0.0;
				segments[ii].caUnits[jj].RyR_SEP = 1.0;
				
				segments[ii].caUnits[jj].Buff_Ca_TnCL = 1.8065e-02;
				if(segments[ii].caUnits[jj].Buff_Ca_TnCL > data->Bmax_TnClow) segments[ii].caUnits[jj].Buff_Ca_TnCL = data->Bmax_TnClow;
				segments[ii].caUnits[jj].Buff_Ca_TnCHc = 1.2774e-1;
				segments[ii].caUnits[jj].Buff_Ca_TnCHm = 5.7527e-3;
				segments[ii].caUnits[jj].Buff_Ca_CaM = 6.9090e-4;
				segments[ii].caUnits[jj].Buff_Ca_Myosin_ca = 3.9052e-3;
				segments[ii].caUnits[jj].Buff_Ca_Myosin_mg = 1.3558e-1;
				if(segments[ii].caUnits[jj].Buff_Ca_Myosin_ca + segments[ii].caUnits[jj].Buff_Ca_Myosin_mg > data->Bmax_myosin) 
				{
					segments[ii].caUnits[jj].Buff_Ca_Myosin_ca = 0;
					segments[ii].caUnits[jj].Buff_Ca_Myosin_mg = 0;
				}
				segments[ii].caUnits[jj].Buff_Ca_SRB = 4.3974e-3;
				segments[ii].caUnits[jj].Buff_Ca_CSQN = data->Bmax_Csqn * Ca_sr / (Ca_sr + (data->koff_csqn / data->kon_csqn)); //1.1352;
				segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt = 0;
				
				segments[ii].caUnits[jj].Buff_Ca_SLLsrs = Buff_Ca_SLLj; 
				segments[ii].caUnits[jj].Buff_Ca_SLHsrs = Buff_Ca_SLHj;
		
				segments[ii].caUnits[jj].Ca_i = Ca_i;
				segments[ii].caUnits[jj].Ca_sr = Ca_sr;
				segments[ii].caUnits[jj].Ca_srs = Ca_srs;
			}
			
			// Other stuff
			segments[ii].K_i = K_i;
			
		}
		
		if(settings->ICaL_Distribution_Scaling == NULL)	settings->initICaLDistribution();
		if(settings->RyR_Distribution_Scaling == NULL)	settings->initRyRDistribution();
		if(settings->NCX_Distribution_Scaling == NULL)	settings->initNCXDistribution();
		if(settings->SERCA_Distribution_Scaling == NULL) settings->initSERCADistribution();
		if(settings->RyRParams_NP == NULL) settings->initRyRParameters();
		
		
		cout << "ICaLDist = [";
		for(int ii=0; ii<settings->N_CaDomains; ii++)
		{
			cout << settings->ICaL_Distribution_Scaling[0][ii] << ",";
		}
		cout << "]" << endl;
		
		cout << "RyRDist = [";
		for(int ii=0; ii<settings->N_CaDomains; ii++)
		{
			cout << settings->RyR_Distribution_Scaling[0][ii] << ",";
		}
		cout << "]" << endl;
		
		cout << "NCXDist = [";
		for(int ii=0; ii<settings->N_CaDomains; ii++)
		{
			cout << settings->NCX_Distribution_Scaling[0][ii] << ",";
		}
		cout << "]" << endl;
		
		cout << "SERCADist = [";
		for(int ii=0; ii<settings->N_CaDomains; ii++)
		{
			cout << settings->SERCA_Distribution_Scaling[0][ii] << ",";
		}
		cout << "]" << endl;
		
		GenPurpose::wait(20.0);
		
	ECl=-log(settings->Cl_o/settings->Cl_i)/data->frt;
}

double HAMM::check_State_Change(double oldval, double dval, double dt)
{
	if(fabs(dval) > Max_State_Change) 
	{
		Max_State_Change = fabs(dval);
	}
	return oldval + dval * dt;
}

void HAMM::update(double t, double dt)
{
	update(0, t, dt, Vm, Vm, Vm, Vm);
} 

void HAMM::update(int cellnumber, double t, double dt, double vleft, double vright)
{
	update(cellnumber, t, dt, vleft, vright, Vm, Vm);
}

void HAMM::update(int cellnumber, double t, double dt, double vleft, double vright, double vup, double vdown)
{
	if(showlog) cout << " t = " << t << ", Vm = " << Vm << ", Cai = " << Ca_i << ", dt = " << dt << endl;
	
	Max_State_Change = 0.0000;

	if(t <= dt)	
	{
		CaT_Min = 1; CaT_Max = -1;
		tmaxdvdt = 0; valmaxdvdt = 0;
		vrest = Vm; vmax = Vm;
		APD = 0;
	}

	// Allow clamping of intracellular ion concentrations when values are set in settings
	// ===================================================================
	if(settings->clamp_K_i >= 0.0) 
	{
		for(int ii=0; ii<settings->N_Segments; ii++)
		{
			segments[ii].K_i = settings->clamp_K_i;
		}
	}
	if(settings->clamp_Na_i >= 0.0)
	{
		for(int ii=0; ii<settings->N_Segments; ii++)
		{
			segments[ii].membrane[0].Na_sl = settings->clamp_Na_i;
			segments[ii].membrane[1].Na_sl = settings->clamp_Na_i;
			
			for(int jj=0; jj<settings->N_CaDomains; jj++)
			{
				segments[ii].caUnits[jj].Na_i = settings->clamp_Na_i;
				segments[ii].caUnits[jj].Na_srs = settings->clamp_Na_i;
			}
		}
	}
	if(settings->clamp_Ca_i >= 0.0)
	{
		for(int ii=0; ii<settings->N_Segments; ii++)
		{
			for(int jj=0; jj<settings->N_CaDomains; jj++)
			{
				segments[ii].caUnits[jj].Ca_i = settings->clamp_Ca_i; 			
				segments[ii].caUnits[jj].Ca_srs = settings->clamp_Ca_i; 			
			}
			
			segments[ii].membrane[0].Ca_sl = settings->clamp_Ca_i;
			segments[ii].membrane[1].Ca_sl = settings->clamp_Ca_i;
		}
	}

    // Determine stimulus current
	if(settings->applyVoltageClamp)
	{
		// Switch to next step in VC protocol
		if(t >= settings->VC_Times[settings->VC_index]) 
		{
			settings->VC_index++;
			Max_State_Change = 100;
		}
		Istim = settings->Istim * (t > 0 && t <= 0+settings->stimdur && settings->stimlocation[cellnumber]);
		if(settings->VC_index > 0) Istim = -50 * (0.005 / dt) * (settings->VC_Values[settings->VC_index-1] - Vm);
	}
	else
	{
		// Current clamp
		Istim = settings->Istim * (t > 0 && t <= 0+settings->stimdur && settings->stimlocation[cellnumber]);
		if(settings->addStimParams == 0)
		{
			Istim += settings->IaddStim * (t > settings->addStimStart && t <= settings->addStimEnd && settings->stimlocationAddStim[cellnumber]);
		}
	}
    
	// Make a copy of concentrations to be able to calculate diffusion
	// ===================================================================		
	for(int ii=0; ii<settings->N_Segments; ii++)
	{
		for(int jj=0; jj<settings->N_CaDomains; jj++)
		{
			Copy_Ca_i[ii][jj] = segments[ii].caUnits[jj].Ca_i;
			Copy_Ca_sr[ii][jj] = segments[ii].caUnits[jj].Ca_sr;
			Copy_Ca_srs[ii][jj] = segments[ii].caUnits[jj].Ca_srs;
			Copy_Na_i[ii][jj] = segments[ii].caUnits[jj].Na_i;
			Copy_Na_srs[ii][jj] = segments[ii].caUnits[jj].Na_srs;
		}
		
		Copy_Ca_sl[ii][0] = segments[ii].membrane[0].Ca_sl;
		Copy_Ca_sl[ii][1] = segments[ii].membrane[1].Ca_sl;
		Copy_Na_sl[ii][0] = segments[ii].membrane[0].Na_sl;
		Copy_Na_sl[ii][1] = segments[ii].membrane[1].Na_sl;
	}
	

	Ca_i = 0; Ca_sl = 0; Ca_junc = 0; Ca_sr = 0; Ca_srs = 0; Na_i = 0; Na_srs = 0; Na_sl = 0;
	INaCa_junc = 0; INaCa_sl = 0; ICab_junc = 0; ICab_sl = 0; IpCa_junc = 0; IpCa_sl = 0; IClCa_junc = 0; IClCa_sl = 0;
	ICaL_K = 0; ICaL_Na_junc = 0; ICaL_Na_sl = 0; ICaL_junc = 0; ICaL_sl = 0;
	Jrel = 0; Jleak = 0; Jup = 0; Buff_Ca_EGTA_cyt = 0;
	
	INa_junc = 0; INa_sl = 0; INaL_junc = 0; INaL_sl = 0; INaK_junc = 0; INaK_sl = 0; INab_junc = 0; INab_sl = 0; IKr = 0; IKs = 0; IK1 = 0; IKp = 0; IKur = 0; ITo1 = 0; IClb = 0; IKAch = 0;
	
	// == Main update loop of all segments
	// ===================================================================	
	for(int ii=0; ii<settings->N_Segments; ii++)
	{
		for(int jj=0; jj<2; jj++)
		{
			if(jj==0) ENa_junc = log(settings->Na_o/segments[ii].caUnits[0].Na_srs)/data->frt;
			if(jj==1) ENa_junc = log(settings->Na_o/segments[ii].caUnits[settings->N_CaDomains-1].Na_srs)/data->frt;

			Ion_HH_Na(ii, jj, dt); 
			Ion_HH_NaL(ii, jj, dt);
			Ion_HH_NaK(ii, jj, dt);
			Ion_HH_Nab(ii, jj, dt);
			Ion_HH_Cl_Background(ii, jj, dt);
			Ion_HH_K1(ii, jj, dt);
			Ion_HH_KAch(ii, jj, dt);
			Ion_HH_Kp(ii, jj, dt);
			Ion_HH_Kur(ii, jj, dt);
			Ion_HH_Kr(ii, jj, dt);
			Ion_HH_To1(ii, jj, dt);
			Ion_HH_Ks(ii, jj, dt);

			Ion_HH_Cl(ii, jj, dt);
			
			Ion_Ca_Handling(ii, jj, dt);     // time-independent calcium currents
		}
	
		// == Calculate all local Ca domains
		// ===================================================================		
		for(int jj=0; jj<settings->N_CaDomains; jj++)
		{
			Ion_Markov_CaL(ii, jj, dt);
			Ion_HH_NaCa(ii, jj, dt);
			
			double Ca_i_L_Domain = segments[ii].caUnits[jj].Ca_i;
			double Ca_sr_L_Domain = segments[ii].caUnits[jj].Ca_sr;
			double Ca_srs_L_Domain = segments[ii].caUnits[jj].Ca_srs;
			double Na_i_L_Domain = segments[ii].caUnits[jj].Na_i;
			double Na_srs_L_Domain = segments[ii].caUnits[jj].Na_srs;
			if(jj > 0) { Ca_i_L_Domain = Copy_Ca_i[ii][jj-1]; Ca_sr_L_Domain = Copy_Ca_sr[ii][jj-1]; Ca_srs_L_Domain = Copy_Ca_srs[ii][jj-1]; Na_i_L_Domain = Copy_Na_i[ii][jj-1]; Na_srs_L_Domain = Copy_Na_srs[ii][jj-1]; }
			
			double Ca_i_R_Domain = segments[ii].caUnits[jj].Ca_i;
			double Ca_sr_R_Domain = segments[ii].caUnits[jj].Ca_sr;
			double Ca_srs_R_Domain = segments[ii].caUnits[jj].Ca_srs;
			double Na_i_R_Domain = segments[ii].caUnits[jj].Na_i;
			double Na_srs_R_Domain = segments[ii].caUnits[jj].Na_srs;
			if(jj < settings->N_CaDomains - 1) { Ca_i_R_Domain = Copy_Ca_i[ii][jj+1]; Ca_sr_R_Domain = Copy_Ca_sr[ii][jj+1]; Ca_srs_R_Domain = Copy_Ca_srs[ii][jj+1]; Na_i_R_Domain = Copy_Na_i[ii][jj+1]; Na_srs_R_Domain = Copy_Na_srs[ii][jj+1]; }
			
			double Ca_i_U_Segment = segments[ii].caUnits[jj].Ca_i;
			double Ca_sr_U_Segment = segments[ii].caUnits[jj].Ca_sr;
			double Ca_srs_U_Segment = segments[ii].caUnits[jj].Ca_srs;
			double Na_i_U_Segment = segments[ii].caUnits[jj].Na_i;
			double Na_srs_U_Segment = segments[ii].caUnits[jj].Na_srs;
			if(ii > 0) { Ca_i_U_Segment = Copy_Ca_i[ii-1][jj]; Ca_sr_U_Segment = Copy_Ca_sr[ii-1][jj]; Ca_srs_U_Segment = Copy_Ca_srs[ii-1][jj]; Na_i_U_Segment = Copy_Na_i[ii-1][jj]; Na_srs_U_Segment = Copy_Na_srs[ii-1][jj]; }
			
			double Ca_i_D_Segment = segments[ii].caUnits[jj].Ca_i;
			double Ca_sr_D_Segment = segments[ii].caUnits[jj].Ca_sr;
			double Ca_srs_D_Segment = segments[ii].caUnits[jj].Ca_srs;
			double Na_i_D_Segment = segments[ii].caUnits[jj].Na_i;
			double Na_srs_D_Segment = segments[ii].caUnits[jj].Na_srs;
			if(ii < settings->N_Segments - 1) { Ca_i_D_Segment = Copy_Ca_i[ii+1][jj]; Ca_sr_D_Segment = Copy_Ca_sr[ii+1][jj]; Ca_srs_D_Segment = Copy_Ca_srs[ii+1][jj]; Na_i_D_Segment = Copy_Na_i[ii+1][jj]; Na_srs_D_Segment = Copy_Na_srs[ii+1][jj]; }

			updateCaDomain(dt, t, ii, jj, Ca_i_L_Domain, Ca_i_R_Domain, Ca_i_U_Segment, Ca_i_D_Segment, Ca_sr_L_Domain, Ca_sr_R_Domain, Ca_sr_U_Segment, Ca_sr_D_Segment, Ca_srs_L_Domain, Ca_srs_R_Domain, Ca_srs_U_Segment, Ca_srs_D_Segment);
			updateNaDomain(dt, t, ii, jj, Na_i_L_Domain, Na_i_R_Domain, Na_i_U_Segment, Na_i_D_Segment, Na_srs_L_Domain, Na_srs_R_Domain, Na_srs_U_Segment, Na_srs_D_Segment);
			
			Ca_i += segments[ii].caUnits[jj].Ca_i;
			Ca_sr += segments[ii].caUnits[jj].Ca_sr;
			Ca_srs += segments[ii].caUnits[jj].Ca_srs;
			Na_i += segments[ii].caUnits[jj].Na_i;
			Na_srs += segments[ii].caUnits[jj].Na_srs;
			Jrel += segments[ii].caUnits[jj].Jrel;
			Jleak += segments[ii].caUnits[jj].Jleak;
			Jup += segments[ii].caUnits[jj].Jup;
			Buff_Ca_EGTA_cyt += segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt;
			
			ICaL_K += segments[ii].caUnits[jj].ICaL_K;
			ICaL_Na_junc += segments[ii].caUnits[jj].ICaL_Na_junc; ICaL_Na_sl += segments[ii].caUnits[jj].ICaL_Na_sl;
			ICaL_junc += segments[ii].caUnits[jj].ICaL_junc; ICaL_sl += segments[ii].caUnits[jj].ICaL_sl;
			INaCa_junc += segments[ii].caUnits[jj].INaCa_junc; INaCa_sl += segments[ii].caUnits[jj].INaCa_sl;
		}
		
		// == Calculate all membrane domains
		// ===================================================================
			for(int jj=0; jj<2; jj++)
			{				
				double Na_sl_U_Segment = segments[ii].membrane[jj].Na_sl;
				double Ca_sl_U_Segment = segments[ii].membrane[jj].Ca_sl;
				if(ii > 0) { Na_sl_U_Segment = Copy_Na_sl[ii-1][jj]; Ca_sl_U_Segment = Copy_Ca_sl[ii-1][jj]; }
				double Na_sl_D_Segment = segments[ii].membrane[jj].Na_sl;
				double Ca_sl_D_Segment = segments[ii].membrane[jj].Ca_sl;
				if(ii < settings->N_Segments - 1) { Na_sl_D_Segment = Copy_Na_sl[ii+1][jj]; Ca_sl_D_Segment = Copy_Ca_sl[ii+1][jj]; }
				double Ca_i_neighbor = Copy_Ca_i[ii][0];			
				if(jj == 1) Ca_i_neighbor = Copy_Ca_i[ii][settings->N_CaDomains-1];
				double Ca_srs_neighbor = Copy_Ca_srs[ii][0];
				if(jj == 1) Ca_srs_neighbor = Copy_Ca_srs[ii][settings->N_CaDomains-1];
				
				double Na_i_neighbor = Copy_Na_i[ii][0];
				if(jj == 1) Na_i_neighbor = Copy_Na_i[ii][settings->N_CaDomains-1];
				
				UpdateConcentrations(ii, jj, dt, t, Na_sl_U_Segment, Na_sl_D_Segment, Ca_sl_U_Segment, Ca_sl_D_Segment, Ca_i_neighbor, Ca_srs_neighbor, Na_i_neighbor);
				
				// Update whole cell currents
				INa_junc += segments[ii].membrane[jj].INa_junc; INa_sl += segments[ii].membrane[jj].INa_sl; 
				INaL_junc += segments[ii].membrane[jj].INaL_junc; INaL_sl += segments[ii].membrane[jj].INaL_sl; 
				INaK_junc += segments[ii].membrane[jj].INaK_junc; INaK_sl += segments[ii].membrane[jj].INaK_sl; 
				INab_junc += segments[ii].membrane[jj].INab_junc; INab_sl += segments[ii].membrane[jj].INab_sl; 
				IKr += segments[ii].membrane[jj].IKr; 
				IKs += segments[ii].membrane[jj].IKs; 
				IK1 += segments[ii].membrane[jj].IK1; 
				IKp += segments[ii].membrane[jj].IKp; 
				IKur += segments[ii].membrane[jj].IKur; 
				ITo1 += segments[ii].membrane[jj].ITo1; 
				IClb += segments[ii].membrane[jj].IClb; 
				IKAch += segments[ii].membrane[jj].IKAch;
				
				Ca_sl += segments[ii].membrane[jj].Ca_sl; Ca_junc += segments[ii].membrane[jj].Ca_junc;
				ICab_junc += segments[ii].membrane[jj].ICab_junc; ICab_sl += segments[ii].membrane[jj].ICab_sl;
				IpCa_junc += segments[ii].membrane[jj].IpCa_junc; IpCa_sl += segments[ii].membrane[jj].IpCa_sl;
				IClCa_junc += segments[ii].membrane[jj].IClCa_junc; IClCa_sl += segments[ii].membrane[jj].IClCa_sl;
			}
			
			double IK_tot = segments[ii].membrane[0].ITo1 + segments[ii].membrane[1].ITo1 + segments[ii].membrane[0].IKr + segments[ii].membrane[1].IKr + segments[ii].membrane[0].IKs + segments[ii].membrane[1].IKs + segments[ii].membrane[0].IK1 + segments[ii].membrane[1].IK1; // + segments[ii].membrane[0].ISK + segments[ii].membrane[1].ISK; 
			//IK_tot = IK_tot - 2*(segments[ii].membrane[0].INaK_junc + segments[ii].membrane[1].INaK_junc + segments[ii].membrane[0].INaK_sl + segments[ii].membrane[1].INaK_sl) + segments[ii].membrane[0].ICaL_K + segments[ii].membrane[1].ICaL_K + segments[ii].membrane[0].IKp + segments[ii].membrane[1].IKp + segments[ii].membrane[0].IKur + segments[ii].membrane[1].IKur + segments[ii].membrane[0].IKAch + segments[ii].membrane[1].IKAch + 2*Istim;
			IK_tot = IK_tot - 2*(segments[ii].membrane[0].INaK_junc + segments[ii].membrane[1].INaK_junc + segments[ii].membrane[0].INaK_sl + segments[ii].membrane[1].INaK_sl) + segments[ii].membrane[0].IKp + segments[ii].membrane[1].IKp + segments[ii].membrane[0].IKur + segments[ii].membrane[1].IKur + segments[ii].membrane[0].IKAch + segments[ii].membrane[1].IKAch + 2*Istim;
			double dK_i = -0.5*IK_tot*data->CmemF/data->vmyo; // 0.0;
			segments[ii].K_i = check_State_Change(segments[ii].K_i, dK_i, dt);
	}		
	
	//Determine whole-cell currents
	INa_junc = INa_junc / (2 * settings->N_Segments); INa_sl = INa_sl / (2 * settings->N_Segments);
	INaL_junc = INaL_junc / (2 * settings->N_Segments); INaL_sl = INaL_sl / (2 * settings->N_Segments);
	INaK_junc = INaK_junc / (2 * settings->N_Segments); INaK_sl = INaK_sl / (2 * settings->N_Segments);
	INab_junc = INab_junc / (2 * settings->N_Segments); INab_sl = INab_sl / (2 * settings->N_Segments);

	IKr = IKr / (2 * settings->N_Segments);
	IKs = IKs / (2 * settings->N_Segments);
	IK1 = IK1 / (2 * settings->N_Segments);
	IKp = IKp / (2 * settings->N_Segments);
	IKur = IKur / (2 * settings->N_Segments);
	ITo1 = ITo1 / (2 * settings->N_Segments);
	IClb = IClb / (2 * settings->N_Segments);
	IKAch = IKAch / (2 * settings->N_Segments);
	
	Ca_junc = Ca_junc / (2 * settings->N_Segments); Ca_sl = Ca_sl / (2 * settings->N_Segments);	
	Na_sl = Na_sl / (2 * settings->N_Segments);
	Na_i = Na_i / (settings->N_Segments * settings->N_CaDomains);
	Na_srs = Na_srs / (settings->N_Segments * settings->N_CaDomains);
	ICab_junc = ICab_junc / (2 * settings->N_Segments); ICab_sl = ICab_sl / (2 * settings->N_Segments);
	IpCa_junc = IpCa_junc / (2 * settings->N_Segments); IpCa_sl = IpCa_sl / (2 * settings->N_Segments);
	IClCa_junc = IClCa_junc / (2 * settings->N_Segments); IClCa_sl = IClCa_sl / (2 * settings->N_Segments);
	
	Ca_i = Ca_i / (settings->N_Segments * settings->N_CaDomains);
	Ca_sr = Ca_sr / (settings->N_Segments * settings->N_CaDomains);
	Ca_srs = Ca_srs / (settings->N_Segments * settings->N_CaDomains);
	Jrel = Jrel / (settings->N_Segments * settings->N_CaDomains);
	Jleak = Jleak / (settings->N_Segments * settings->N_CaDomains);
	Jup = Jup / (settings->N_Segments * settings->N_CaDomains);
	Buff_Ca_EGTA_cyt = Buff_Ca_EGTA_cyt / (settings->N_Segments * settings->N_CaDomains);
	
	// == Update Vm and global ion concentrations
	// ===================================================================	
		caiont = (ICaL_junc+ICaL_sl)+(ICab_junc+ICab_sl)+(IpCa_junc+IpCa_sl)-2*(INaCa_junc+INaCa_sl);
		naiont = (ICaL_Na_junc+ICaL_Na_sl)+(INa_junc+INa_sl)+3*(INaCa_junc+INaCa_sl)+3*(INaK_junc+INaK_sl)+(INaL_junc+INaL_sl)+(INab_junc+INab_sl);
		kiont = ICaL_K+IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1+IKAch+Istim;//+ISK;
		clont = IClb+(IClCa_junc+IClCa_sl);
		
		double Rgap = settings->Rgap;
		if(cellnumber >= settings->strandLowConductionStart && cellnumber <= settings->strandLowConductionEnd) Rgap = Rgap * settings->strandLowConductionFactor;		
		double RgapupDown = Rgap * settings->CVAnisotropyFactor;
		double loc_l = data->l / 1.0E4; // conversion to cm
		double loc_a = data->a / 1.0E4;		
		
		// New 2D, more optimized, correct units (note sqrt(PI) = 1.772453, RCG = 2, dx = loc_l, dy = 2 * loc_a)
		double dvdt = 1000 * ((vleft - 2*Vm + vright) / ((data->Rmyo + Rgap / loc_l) * 4 * loc_l * loc_l / loc_a) + (vup - 2*Vm + vdown) / ((data->Rmyo + RgapupDown / (2 * loc_a)) * 8 * loc_a * 1.77245)) - (naiont+kiont+caiont+clont);		
		Vm = check_State_Change(Vm, dvdt, dt);

		if(Ca_i < CaT_Min) CaT_Min = Ca_i;
		if(Ca_i > CaT_Max) CaT_Max = Ca_i;
	
		if(dvdt > valmaxdvdt)
		{
			valmaxdvdt = dvdt;
			tmaxdvdt = t;
		}
		else
		{
			if(t < tmaxdvdt + 4 && Vm > vmax) vmax = Vm;
		}
		if(vmax > -50 && t > tmaxdvdt + 4 && Vm < (vrest + (1 - settings->APDRepLevel) * (vmax - vrest)) && APD == 0) APD = t - tmaxdvdt;
}

void HAMM::Ion_HH_Na(int seg_i, int mem_i, double dt)
{
	if(fabs(settings->INaB - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].INa_junc = 0;
		segments[seg_i].membrane[mem_i].INa_sl = 0;
		return;
	}
	
	double ENa_sl = log(settings->Na_o/segments[seg_i].membrane[mem_i].Na_sl)/data->frt;
	double Po_Na = (1 - settings->INaB) * data->GNa * segments[seg_i].membrane[mem_i].m_Na*segments[seg_i].membrane[mem_i].m_Na*segments[seg_i].membrane[mem_i].m_Na * segments[seg_i].membrane[mem_i].h_Na * segments[seg_i].membrane[mem_i].j_Na;
	segments[seg_i].membrane[mem_i].INa_junc = data->Fjunc * Po_Na * (Vm-ENa_junc);
	segments[seg_i].membrane[mem_i].INa_sl = (1-data->Fjunc) * Po_Na * (Vm-ENa_sl);
	
	double mss, taum, hss, tauh, jss, tauj;
	if(!settings->useCourtemancheINa)
	{
		// Grandi et al. Circ Res. 2011 model of INa		
		double mss_temp = 1 + exp( -(56.86 + Vm) / 9.03 );
		mss = 1 / (mss_temp * mss_temp);
		taum = 0.1292 * exp(-((Vm+45.79)/15.54)*((Vm+45.79)/15.54)) + 0.06487 * exp(-((Vm-4.823)/51.12)*((Vm-4.823)/51.12));                 
 
		double ah = 0;
		if(Vm < -40) ah = (0.057 * exp( -(Vm + 80) / 6.8 )); 	
		double bh = (0.77 / (0.13*(1 + exp( -(Vm + 10.66) / 11.1 ))));
		if(Vm < -40) bh = ((2.7 * exp( 0.079 * Vm) + 3.1E5 * exp(0.3485 * Vm))); 
	
		double hss_temp = 1 + exp( (Vm + 71.55)/7.43);
		tauh = 1 / (ah + bh); 
		hss = 1 / (hss_temp*hss_temp);
 
		double aj = 0;
		if(Vm < -40) aj = (((-2.5428E4*exp(0.2444*Vm) - 6.948E-6 * exp(-0.04391*Vm)) * (Vm + 37.78)) / (1 + exp( 0.311 * (Vm + 79.23) )));	
		double bj = ((0.6 * exp( 0.057 * Vm)) / (1 + exp( -0.1 * (Vm + 32) )));
		if(Vm < -40) bj = ((0.02424 * exp( -0.01052 * Vm )) / (1 + exp( -0.1378 * (Vm + 40.14) ))); 
	
		tauj = 1 / (aj + bj);
		jss = hss; // 1 / ((1 + exp( (Vm + 71.55)/7.43 )*(1 + exp( (Vm + 71.55)/7.43 ));
	}
	else
	{
		// Courtemanche et al. model
		double am = 0.32 * (Vm + 47.13) / (1 - exp(-0.1 * (Vm + 47.13)));
		if(Vm == 47.13) am = 3.2;
		double bm = 0.08 * exp(-Vm / 11);
		mss = am / (am + bm);
		taum = 1 / (am + bm); 
		
		double ah = 0;
		if(Vm < -40) ah = (0.135 * exp( -(Vm + 80) / 6.8 )); 	
		double bh = (1.0 / (0.13*(1 + exp( -(Vm + 10.66) / 11.1 ))));
		if(Vm < -40) bh = 3.56 * exp( 0.079 * Vm) + 3.1E5 * exp(0.35 * Vm);
		hss = ah / (ah + bh);
		tauh = 1 / (ah + bh); 
		
		double aj = 0;
		if(Vm < -40) aj = (((-127140*exp(0.2444*Vm) - 3.474E-5 * exp(-0.04391*Vm)) * (Vm + 37.78)) / (1 + exp( 0.311 * (Vm + 79.23) )));	
		double bj = ((0.3 * exp(-2.5535E-7 * Vm)) / (1 + exp( -0.1 * (Vm + 32) )));
		if(Vm < -40) bj = ((0.1212 * exp( -0.01052 * Vm )) / (1 + exp( -0.1378 * (Vm + 40.14) ))); 
		jss = aj / (aj + bj);
		tauj = 1 / (aj + bj); 
	}
	
	segments[seg_i].membrane[mem_i].m_Na = check_State_Change(segments[seg_i].membrane[mem_i].m_Na, (mss - segments[seg_i].membrane[mem_i].m_Na) / taum, dt); 
		if(segments[seg_i].membrane[mem_i].m_Na < 0.0) segments[seg_i].membrane[mem_i].m_Na = 0.0;
		if(segments[seg_i].membrane[mem_i].m_Na > 1.0) segments[seg_i].membrane[mem_i].m_Na = 1.0;
		
	segments[seg_i].membrane[mem_i].h_Na = check_State_Change(segments[seg_i].membrane[mem_i].h_Na, (hss - segments[seg_i].membrane[mem_i].h_Na) / tauh, dt);
	segments[seg_i].membrane[mem_i].j_Na = check_State_Change(segments[seg_i].membrane[mem_i].j_Na, (jss - segments[seg_i].membrane[mem_i].j_Na) / tauj, dt);	
}

void HAMM::Ion_HH_NaL(int seg_i, int mem_i, double dt)
{
	if(data->isAF && fabs(settings->INaLB - 1.0) > 1E-10)
	{
		double GNaL=0.0025;
		double aml = 0.32*(Vm+47.13)/(1-exp(-0.1*(Vm+47.13)));
		double bml = 0.08*exp(-Vm/11);
		double hlinf = 1/(1+exp((Vm+91)/6.1));
		double tauhl=600;
		
		segments[seg_i].membrane[mem_i].mL_Na = check_State_Change(segments[seg_i].membrane[mem_i].mL_Na, aml * (1-segments[seg_i].membrane[mem_i].mL_Na) - bml * segments[seg_i].membrane[mem_i].mL_Na, dt);
			if(segments[seg_i].membrane[mem_i].mL_Na < 0.0) segments[seg_i].membrane[mem_i].mL_Na = 0.0;
			if(segments[seg_i].membrane[mem_i].mL_Na > 1.0) segments[seg_i].membrane[mem_i].mL_Na = 1.0;
		segments[seg_i].membrane[mem_i].hL_Na = check_State_Change(segments[seg_i].membrane[mem_i].hL_Na, (hlinf - segments[seg_i].membrane[mem_i].hL_Na)/tauhl, dt);

		double ENa_sl = log(settings->Na_o/segments[seg_i].membrane[mem_i].Na_sl)/data->frt;
		segments[seg_i].membrane[mem_i].INaL_junc = (1 - settings->INaLB) * data->Fjunc * GNaL * segments[seg_i].membrane[mem_i].mL_Na*segments[seg_i].membrane[mem_i].mL_Na*segments[seg_i].membrane[mem_i].mL_Na * segments[seg_i].membrane[mem_i].hL_Na * (Vm-ENa_junc);
		segments[seg_i].membrane[mem_i].INaL_sl = (1 - settings->INaLB) * (1-data->Fjunc) * GNaL * segments[seg_i].membrane[mem_i].mL_Na*segments[seg_i].membrane[mem_i].mL_Na*segments[seg_i].membrane[mem_i].mL_Na * segments[seg_i].membrane[mem_i].hL_Na * (Vm-ENa_sl);
	}
	else
	{
		segments[seg_i].membrane[mem_i].INaL_junc = 0.0;
		segments[seg_i].membrane[mem_i].INaL_sl = 0.0;
	}
}

void HAMM::Ion_HH_NaCa(int seg_i, int cadom_i, double dt)
{	
	if(settings->clamp_NCX > -9.9)
	{
		segments[seg_i].caUnits[cadom_i].INaCa_junc = settings->clamp_NCX;
		segments[seg_i].caUnits[cadom_i].INaCa_sl = settings->clamp_NCX;
	}
	else
	{
		if(settings->Ca_o > 0 && settings->NCX_Distribution_Scaling[seg_i][cadom_i]>1E-10)
		{
			double Ca_junc_observed = segments[seg_i].caUnits[cadom_i].Ca_srs;
			double Ca_sl_observed = segments[seg_i].caUnits[cadom_i].Ca_i;
			if(cadom_i == 0) Ca_sl_observed = segments[seg_i].membrane[0].Ca_sl;
			if(cadom_i == settings->N_CaDomains-1) Ca_sl_observed = segments[seg_i].membrane[1].Ca_sl;
			
			double Na_junc_observed = segments[seg_i].caUnits[cadom_i].Na_srs;
			double Na_sl_observed = segments[seg_i].caUnits[cadom_i].Na_i;
			if(cadom_i == 0) Na_sl_observed = segments[seg_i].membrane[0].Na_sl;
			if(cadom_i == settings->N_CaDomains-1) Na_sl_observed = segments[seg_i].membrane[1].Na_sl;
			
			double Na_o_p3 = settings->Na_o * settings->Na_o * settings->Na_o;
			double Na_sl_p3 =  Na_sl_observed * Na_sl_observed * Na_sl_observed;
			double Na_junc_p3 =  Na_junc_observed * Na_junc_observed * Na_junc_observed;
			
			double Ka_junc = 1 / (1+pow(data->Kdact/Ca_junc_observed, 2));
			double Ka_sl = 1 / (1+pow(data->Kdact/Ca_sl_observed, 2));

				double Vm_s1 = exp(data->nu*Vm*data->frt) * settings->Ca_o;
				double s1_junc = Vm_s1 * Na_junc_p3;
				double s1_sl = Vm_s1 * Na_sl_p3;
				double Vm_s2 = exp((data->nu-1)*Vm*data->frt);
				double s2_junc = Vm_s2 * Na_o_p3 * Ca_junc_observed;
				double s3_junc = data->KmCai*Na_o_p3*(1+pow(Na_junc_observed/data->KmNai,3)) + data->KmNao_p3*Ca_junc_observed*(1+Ca_junc_observed/data->KmCai) + data->KmCao*Na_junc_p3 + Na_junc_p3*settings->Ca_o + Na_o_p3*Ca_junc_observed;
				double s2_sl = Vm_s2 * Na_o_p3 * Ca_sl_observed;
				double s3_sl = data->KmCai*Na_o_p3*(1+pow(Na_sl_observed/data->KmNai,3)) + data->KmNao_p3*Ca_sl_observed*(1+Ca_sl_observed/data->KmCai) + data->KmCao*Na_sl_p3 + Na_sl_p3*settings->Ca_o + Na_o_p3*Ca_sl_observed;

				segments[seg_i].caUnits[cadom_i].INaCa_junc = (1 - settings->INaCaB) * data->Fjunc * settings->NCX_Distribution_Scaling[seg_i][cadom_i] * data->IbarNCX * Ka_junc * (s1_junc-s2_junc)/s3_junc/(1+data->ksat*Vm_s2);
				segments[seg_i].caUnits[cadom_i].INaCa_sl = (1 - settings->INaCaB) * (1 - data->Fjunc) * settings->NCX_Distribution_Scaling[seg_i][cadom_i] * data->IbarNCX * Ka_sl * (s1_sl-s2_sl)/s3_sl/(1+data->ksat*Vm_s2);
		}
		else
		{
			segments[seg_i].caUnits[cadom_i].INaCa_junc = 0;
			segments[seg_i].caUnits[cadom_i].INaCa_sl = 0;
		}
	}
}

void HAMM::Ion_HH_NaK(int seg_i, int mem_i, double dt)
{ 
	if(seg_i == 0 && mem_i == 0)
	{
		// Only need to calculate these once every time step
		sigma = (exp(settings->Na_o/67.3)-1)/7;
		fnak = 1 / (1 + 0.1245*exp(-0.1*Vm*data->frt) + 0.0365*sigma*exp(-Vm*data->frt));
	}
	
	double INaK_junc_NP;
	if(mem_i == 0) INaK_junc_NP = data->Fjunc * data->IbarNaK * fnak * settings->K_o / (1+pow(data->KmNaip_NP/segments[seg_i].caUnits[0].Na_srs,4)) / (settings->K_o + data->KmKo);
	if(mem_i == 1) INaK_junc_NP = data->Fjunc * data->IbarNaK * fnak * settings->K_o / (1+pow(data->KmNaip_NP/segments[seg_i].caUnits[settings->N_CaDomains-1].Na_srs,4)) / (settings->K_o + data->KmKo);
	double INaK_sl_NP = (1.0-data->Fjunc) * data->IbarNaK * fnak * settings->K_o / (1+pow(data->KmNaip_NP/segments[seg_i].membrane[mem_i].Na_sl,4)) / (settings->K_o + data->KmKo);
	
	segments[seg_i].membrane[mem_i].INaK_junc = (1 - settings->INaKB) * INaK_junc_NP; 
	segments[seg_i].membrane[mem_i].INaK_sl = (1 - settings->INaKB) * INaK_sl_NP;
}

void HAMM::Ion_HH_Nab(int seg_i, int mem_i, double dt)
{
	double ENa_sl = log(settings->Na_o/segments[seg_i].membrane[mem_i].Na_sl)/data->frt;
	segments[seg_i].membrane[mem_i].INab_junc = (1 - settings->INabB) * data->Fjunc * data->GNab * (Vm-ENa_junc);
	segments[seg_i].membrane[mem_i].INab_sl = (1 - settings->INabB) * (1-data->Fjunc) * data->GNab * (Vm-ENa_sl);
}

// =======================================================================
// IK1 with [Na+]i-dependent regulation
// Formulation from Voigt, Heijman, et al. J. Mol. Cell. Cardiol. 2013
// =======================================================================
void HAMM::Ion_HH_K1(int seg_i, int mem_i, double dt)
{
	if(fabs(settings->IK1B - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].IK1 = 0;
		return;
	}
	
	double Na_observed = segments[seg_i].membrane[mem_i].Na_sl;
	if(settings->clampIK1NaDependence >= 0) Na_observed = settings->clampIK1NaDependence;
	
	double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
	double ak1 = (0.1 + 0.9 / (1 + ((Na_observed / 7.0)*(Na_observed / 7.0)))) * 1.0 / (1+exp(0.2385*(Vm-EK-59.215)));
	double GK1 = 2.1 * data->GK1max * sqrt(settings->K_o/5.4);
	
    double bk1 = (0.49124*exp(0.08032*(Vm-EK+5.476))+exp(0.06175*(Vm-EK-594.31)))/(1+exp(-0.5143*(Vm-EK+4.753)));
	double k1ss = ak1/(ak1+bk1);
    
	segments[seg_i].membrane[mem_i].IK1 = (1 - settings->IK1B) * GK1 * k1ss * (Vm-EK);
}

// =======================================================================
// IK,ACh with [Na+]i-dependent regulation
// Formulation from Voigt, Heijman, et al. J. Mol. Cell. Cardiol. 2013
// =======================================================================
void HAMM::Ion_HH_KAch(int seg_i, int mem_i, double dt)
{
	segments[seg_i].membrane[mem_i].IKAch = 0.0;
	if(settings->simIKAch)
	{
		double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
		double Na_observed = segments[seg_i].membrane[mem_i].Na_sl;
		if(settings->clampIKAChNaDependence >= 0) Na_observed = settings->clampIKAChNaDependence;
		if(Na_observed == 0) Na_observed = 0.0001; // Numerical issues when exactly zero
		
		// settings->simIKAChNaDep = 1 --> Forced Na+-dependence in both SR and cAF (even though none measured in cAF)
		// settings->simIKAChNaDep = 0 --> Forced absence of Na+-dependence in both SR and cAF 
		// settings->simIKAChNaDep = -1 --> Na+-dependence only in Ctl
		if(!data->isAF)
		{
			if(settings->simIKAChNaDep == -1 || settings->simIKAChNaDep == 1)
			{
				segments[seg_i].membrane[mem_i].IKAch = (1 - settings->IKAchB) * 0.10 * (1 + 1.50 / (1 + pow(9.0 / Na_observed,4))) * (0.055 + 0.40 / (1 + exp((Vm -EK + 9.53) / 17.18))) * (Vm - EK) * sqrt(settings->K_o/5.4);
			}
			else
			{
				segments[seg_i].membrane[mem_i].IKAch = (1 - settings->IKAchB) * 0.10 * (0.055 + 0.40 / (1 + exp((Vm -EK + 9.53) / 17.18))) * (Vm - EK) * sqrt(settings->K_o/5.4);
			}			
		}
		else
		{
			if(settings->simIKAChNaDep == -1 || settings->simIKAChNaDep == 0)
			{
				segments[seg_i].membrane[mem_i].IKAch = (1 - settings->IKAchB) * 0.05 * (0.055 + 0.40 / (1 + exp((Vm -EK + 9.53) / 17.18))) * (Vm - EK) * sqrt(settings->K_o/5.4);
			}
			else
			{
				segments[seg_i].membrane[mem_i].IKAch = (1 - settings->IKAchB) * 0.05 * (1 + 1.5 / (1 + pow(9.0 / Na_observed,4))) * (0.055 + 0.40 / (1 + exp((Vm -EK + 9.53) / 17.18))) * (Vm - EK) * sqrt(settings->K_o/5.4);
			}
		}
	}
}

void HAMM::Ion_HH_Kp(int seg_i, int mem_i, double dt)
{
	if(fabs(settings->IKpB - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].IKp = 0;
		return;
	}
	
	double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
	segments[seg_i].membrane[mem_i].IKp = (1 - settings->IKpB) * data->GKp * (Vm - EK) / (1 + exp((7.488 - Vm) / 5.98));
}

void HAMM::Ion_HH_Kur(int seg_i, int mem_i, double dt)
{	
	if(fabs(settings->IKurB - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].IKur = 0;
		return;
	}
	
	if(seg_i == 0 && mem_i == 0)
	{
		// Only need to be calculated once every time step.
		
		// Activation
		xkurss = 1 / (1 + exp(-(Vm+6)/8.6));
		tauxkur = 9/(1+exp((Vm+5)/12.0))+0.5;

		// Inactivation
		ykurss = 1 / (1 + exp((Vm+7.5)/10));
		tauykur = 590 / (1+exp((Vm+60)/10.0))+3050;
	}

	double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
	segments[seg_i].membrane[mem_i].IKur = (1 - settings->IKurB) * data->GKur * segments[seg_i].membrane[mem_i].kur_a * segments[seg_i].membrane[mem_i].kur_i * (Vm-EK);
	
	segments[seg_i].membrane[mem_i].kur_a = check_State_Change(segments[seg_i].membrane[mem_i].kur_a, (xkurss - segments[seg_i].membrane[mem_i].kur_a) / tauxkur, dt);	
	segments[seg_i].membrane[mem_i].kur_i = check_State_Change(segments[seg_i].membrane[mem_i].kur_i, (ykurss - segments[seg_i].membrane[mem_i].kur_i) / tauykur, dt);
}

void HAMM::Ion_HH_Kr(int seg_i, int mem_i, double dt)
{   	
	if(fabs(settings->IKrB - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].IKr = 0;
		return;
	}
	
	if(seg_i==0 && mem_i==0)
	{
		// Only need to be calculated once every time step.
		gkr = data->GKrmax*sqrt(settings->K_o/5.4);
		ikr_r = 1/(1+exp((Vm+74)/24.0));
	}
	
	double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
    segments[seg_i].membrane[mem_i].IKr = (1 - settings->IKrB) *gkr*segments[seg_i].membrane[mem_i].xr*ikr_r*(Vm-EK);
	
	double xrss = 1/(1+exp(-(Vm+10.0)/5.0));
	double tauxr = 550 / (1+exp((-22-Vm)/9.0)) * 6 / (1+exp((Vm+11)/9)) + 230 / (1+exp((Vm+40)/20));
	segments[seg_i].membrane[mem_i].xr = check_State_Change(segments[seg_i].membrane[mem_i].xr, (xrss - segments[seg_i].membrane[mem_i].xr) / tauxr, dt);	
}

void HAMM::Ion_HH_To1(int seg_i, int mem_i, double dt)
{
	if(fabs(settings->ITo1B - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].ITo1 = 0;
		return;
	}
	
	if(seg_i == 0 && mem_i == 0)
	{
		// Only need to be calculated once every time step.
		
		// Activation
		xtoss = 1.0 / (1 + exp(-(Vm+1.0)/11.0));
		tauxtof = 3.5*exp(-((Vm/30.0)*(Vm/30.0)))+1.5;

		// Equations for Inactivation;
		ytoss = 1.0 / ( 1 + exp((Vm+40.5)/11.5));
		tauytof = 25.635*exp(-((Vm+52.45)/15.8827)*((Vm+52.45)/15.8827))+24.14;
	}

	double EK = log(settings->K_o/segments[seg_i].K_i)/data->frt;
	segments[seg_i].membrane[mem_i].ITo1 = (1 - settings->ITo1B) * data->GToFast * segments[seg_i].membrane[mem_i].to_a * segments[seg_i].membrane[mem_i].to_if * (Vm - EK);
	
	segments[seg_i].membrane[mem_i].to_a = check_State_Change(segments[seg_i].membrane[mem_i].to_a, (xtoss - segments[seg_i].membrane[mem_i].to_a) / tauxtof, dt);
	segments[seg_i].membrane[mem_i].to_if = check_State_Change(segments[seg_i].membrane[mem_i].to_if, (ytoss - segments[seg_i].membrane[mem_i].to_if) / tauytof, dt);
}

void HAMM::Ion_HH_Cl_Background(int seg_i, int mem_i, double dt)
{   
	segments[seg_i].membrane[mem_i].IClb = (1 - settings->IClbB) * data->GClb*(Vm-ECl);
}

void HAMM::Ion_HH_Cl(int seg_i, int mem_i, double dt)
{
	double Ca_junc_observed;
	if(mem_i == 0) Ca_junc_observed = segments[seg_i].caUnits[0].Ca_srs;
	if(mem_i == 1) Ca_junc_observed = segments[seg_i].caUnits[settings->N_CaDomains-1].Ca_srs;
		
	segments[seg_i].membrane[mem_i].IClCa_junc = (1 - settings->IClCaB) * data->Fjunc * data->GClCa / (1+data->KdClCa/Ca_junc_observed)*(Vm-ECl);
	segments[seg_i].membrane[mem_i].IClCa_sl = (1 - settings->IClCaB) * (1 - data->Fjunc) * data->GClCa / (1+data->KdClCa/segments[seg_i].membrane[mem_i].Ca_sl)*(Vm-ECl);
}

void HAMM::Ion_HH_Ks(int seg_i, int mem_i, double dt)
{	
	if(fabs(settings->IKsB - 1.0) < 1E-10)
	{
		segments[seg_i].membrane[mem_i].IKs = 0;
		return;
	}
	
	double EKs = (1/data->frt) * log((settings->K_o + data->pNaK * settings->Na_o)/(K_i + data->pNaK * segments[seg_i].membrane[mem_i].Na_sl));
	double IKs_NP = (1 - settings->IKsB) * data->GKs_NP * segments[seg_i].membrane[mem_i].xs_NP * segments[seg_i].membrane[mem_i].xs_NP * (Vm-EKs);
	segments[seg_i].membrane[mem_i].IKs = IKs_NP;
	
	if(seg_i == 0 && mem_i == 0)
	{
		xsss_NP = 1 / (1+exp(-(Vm+3.8)/14.25)); 
		tauxs_NP =990.1/(1+exp(-(Vm+2.436)/14.12));	
	}
	segments[seg_i].membrane[mem_i].xs_NP = check_State_Change(segments[seg_i].membrane[mem_i].xs_NP, (xsss_NP - segments[seg_i].membrane[mem_i].xs_NP) / tauxs_NP, dt);
}

// =======================================================================
// ICa,L Markov Model to simulate Vm and Ca2+-dependent inactivation
// Formulation from Decker, Heijman, et al. Am. J. Physiol. Heart Circ. Physiol. 2009
// =======================================================================
void HAMM::Ion_Markov_CaL(int seg_i, int cadom_i, double dt)
{
	double* params = settings->Params_ICaL;
	double fact_AF = 1.0;
	if(settings->clamp_ICaL_CDI < 0) settings->clamp_ICaL_CDI = 1; //This means no clamping
		
	double Ca_junc_observed = segments[seg_i].caUnits[cadom_i].Ca_srs;
		
	// NP Currents	
	if(settings->Ca_o > 0 && settings->ICaL_Distribution_Scaling[seg_i][cadom_i] > 1E-10)
	{
		double exp_Vm_frt = exp(Vm*data->frt);
		double exp_2_Vm_frt = exp_Vm_frt * exp_Vm_frt;
		double ibarca_j = settings->ICaL_Distribution_Scaling[seg_i][cadom_i] * fact_AF * params[21]*4*(Vm*data->F*data->frt) * (0.341*Ca_junc_observed*exp_2_Vm_frt-0.341*settings->Ca_o) / (exp_2_Vm_frt-1);
		double ibark = settings->ICaL_Distribution_Scaling[seg_i][cadom_i] * fact_AF * data->p_ICaL_K*(Vm*data->F*data->frt) * (0.75*segments[seg_i].K_i*exp_Vm_frt-0.75*settings->K_o) / (exp_Vm_frt-1);
		double ibarna_j = settings->ICaL_Distribution_Scaling[seg_i][cadom_i] * fact_AF * data->p_ICaL_Na*(Vm*data->F*data->frt) * (0.75*segments[seg_i].caUnits[cadom_i].Na_srs*exp_Vm_frt-0.75*settings->Na_o) / (exp_Vm_frt-1);
		
		double ICaL_junc_NP = (1 - settings->ICaLB) * ibarca_j * (segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + segments[seg_i].caUnits[cadom_i].ICaLState_NP[3]);
		double ICaL_sl_NP = 0;
		double ICaL_K_NP = (1 - settings->ICaLB) * ibark * (segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + segments[seg_i].caUnits[cadom_i].ICaLState_NP[3]);
		double ICaL_Na_junc_NP = (1 - settings->ICaLB) * ibarna_j * (segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + segments[seg_i].caUnits[cadom_i].ICaLState_NP[3]);
		double ICaL_Na_sl_NP = 0;
	
		//Total currents
		segments[seg_i].caUnits[cadom_i].ICaL_junc = ICaL_junc_NP;
		segments[seg_i].caUnits[cadom_i].ICaL_sl = ICaL_sl_NP;
		segments[seg_i].caUnits[cadom_i].ICaL = segments[seg_i].caUnits[cadom_i].ICaL_junc + segments[seg_i].caUnits[cadom_i].ICaL_sl;
		segments[seg_i].caUnits[cadom_i].ICaL_Na_junc = ICaL_Na_junc_NP;
		segments[seg_i].caUnits[cadom_i].ICaL_Na_sl = ICaL_Na_sl_NP;
		segments[seg_i].caUnits[cadom_i].ICaL_Na = segments[seg_i].caUnits[cadom_i].ICaL_Na_junc + segments[seg_i].caUnits[cadom_i].ICaL_Na_sl;
		segments[seg_i].caUnits[cadom_i].ICaL_K = ICaL_K_NP;
		segments[seg_i].caUnits[cadom_i].ICaL_tot = segments[seg_i].caUnits[cadom_i].ICaL_K + segments[seg_i].caUnits[cadom_i].ICaL_Na + segments[seg_i].caUnits[cadom_i].ICaL;
	}
	else
	{
		segments[seg_i].caUnits[cadom_i].ICaL_junc = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_sl = 0;
		segments[seg_i].caUnits[cadom_i].ICaL = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_Na_junc = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_Na_sl = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_Na = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_K = 0;
		segments[seg_i].caUnits[cadom_i].ICaL_tot = 0;
	}
	
	if(settings->ICaL_Distribution_Scaling[seg_i][cadom_i] > 1E-10)
	{
		double ICaL_State_Sum;
	 
		//Determine NP rates
		double ACT_tau, ACT_ss, I_Vbeta_expfunc, Boltz_inact, ACT_ss_steepness;

		ACT_tau = (params[0] + (params[1] * exp(params[2] * (Vm + params[3]))) / (1 + exp(params[4] * (Vm + params[3]))));
		ACT_ss = 1 / ((1 + exp( - (Vm - params[5]) / (params[6]))) * (1 + exp( - (Vm + 25) / 5)));
		 
		I_Vbeta_expfunc = (1 + exp( -(Vm + params[11]) / params[12] ));
		Boltz_inact = (1 / (1 + exp( (Vm + params[13]) / params[14] )));
		
		double I_Valpha = 1 / (params[7] * (1 + exp( ( Vm + params[8] ) / params[9])));
		double I_Vbeta = 1 / (params[10] * I_Vbeta_expfunc);
		double I_Vss = (1 / (1 + params[15])) * (params[15] + Boltz_inact);	
		double I_Vtau = (1 / (I_Valpha + I_Vbeta));
		if(I_Vtau < 0.1) I_Vtau = 0.1;
		
		double Is_Valpha = I_Valpha;	
		double p_Ca_Aff_4 = (params[22] / Ca_junc_observed) * (params[22] / Ca_junc_observed); // * (params[22] / Ca_junc_observed) * (params[22] / Ca_junc_observed); //pow(params[22] / Ca_observed,4)
		double Is_Vbeta_ss = params[16] - settings->clamp_ICaL_CDI * ( params[17] / (1 + p_Ca_Aff_4));
		double Is_Vbeta = 1 / (Is_Vbeta_ss * I_Vbeta_expfunc);
		double Is_Vss = (1 / (1 + params[18])) * (params[18] + Boltz_inact);
		double Is_Vtau = (1 / (Is_Valpha + Is_Vbeta));
		if(Is_Vtau < 0.1) Is_Vtau = 0.1;
		  
		if(settings->clamp_ICaL_VDI >= 0 && settings->clamp_ICaL_VDI <= 1.0)
		{
			I_Vss = settings->clamp_ICaL_VDI;
			Is_Vss = settings->clamp_ICaL_VDI;
		}
		
		double alpha = ACT_ss / ACT_tau;
		double beta = (1 - ACT_ss) / ACT_tau;
		double x = I_Vss / I_Vtau; 
		double y = (1 - I_Vss) / I_Vtau;
		double xs = Is_Vss / Is_Vtau;
		double ys = (1 - Is_Vss) / Is_Vtau;
		double delta = settings->clamp_ICaL_CDI * params[23] * (params[19] / (1 + p_Ca_Aff_4));
		double theta = params[23] * 1;
		double theta_I = params[20];  
		double delta_I = theta_I;
		if(fabs(y) > 1E-20) delta_I = theta_I * ( (x * ys * delta) / (y * xs * theta) );
	 
		double dt_ICaL_C = ( beta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + theta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] + x * segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] - (alpha + delta + y) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] ); 
		double dt_ICaL_O = ( alpha * segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] + theta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] + x * segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] - (beta + delta + y) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] ); 
		double dt_ICaL_Cs = ( delta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] + beta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] + xs * segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] - (theta + alpha + ys) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] ); 
		double dt_ICaL_Os = ( delta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + alpha * segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] + xs * segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] - (theta + beta + ys) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] ); 
		double dt_ICaL_CI = ( y * segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] + beta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] + theta_I * segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] - (x + alpha + delta_I) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] ); 
		double dt_ICaL_OI = ( y * segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] + alpha * segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] + theta_I * segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] - (x + beta + delta_I) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] ); 
		double dt_ICaL_CIs = ( ys * segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] + delta_I * segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] + beta * segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] - (xs + theta_I + alpha) * segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] ); 
		//double dt_ICaL_OIs = ( ys * segments[seg_i].membrane[mem_i].ICaLState_NP[3] + delta_I * segments[seg_i].membrane[mem_i].ICaLState_NP[5] + alpha * segments[seg_i].membrane[mem_i].ICaLState_NP[6] - (xs + theta_I + beta) * segments[seg_i].membrane[mem_i].ICaLState_NP[7] ); 
	 
		segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[0], dt_ICaL_C, dt); segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[1], dt_ICaL_O, dt); segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[2], dt_ICaL_Cs, dt); segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[3], dt_ICaL_Os, dt);
		segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[4], dt_ICaL_CI, dt); segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[5], dt_ICaL_OI, dt); segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] = check_State_Change(segments[seg_i].caUnits[cadom_i].ICaLState_NP[6], dt_ICaL_CIs, dt); 
		segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] = 1.0 - segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] - segments[seg_i].caUnits[cadom_i].ICaLState_NP[6];
	  
		// Show error message when one of the states becomes negative
		if(segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] < -1E-6 || segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] < -1E-6)
		{
			cout << " ICaLState= [" << segments[seg_i].caUnits[cadom_i].ICaLState_NP[0] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[1] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[2] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[3] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] << "," << segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] << "]" << endl;
			cout << " alpha = " << alpha << ", beta = " << beta << ", x = " << x << ", y = " << y << ", xs = " << xs << ", ys = " << ys << ", delta = " << delta << ", theta = " << theta << ", delta_I = " << delta_I << ", theta_I = " << theta_I << endl;
		}
		if(segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] < 0) {segments[seg_i].caUnits[cadom_i].ICaLState_NP[6] += segments[seg_i].caUnits[cadom_i].ICaLState_NP[4]; segments[seg_i].caUnits[cadom_i].ICaLState_NP[4] = 0; }
		if(segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] < 0) {segments[seg_i].caUnits[cadom_i].ICaLState_NP[7] += segments[seg_i].caUnits[cadom_i].ICaLState_NP[5]; segments[seg_i].caUnits[cadom_i].ICaLState_NP[5] = 0; }  
	}
}

void HAMM::Ion_Ca_Handling(int seg_i, int mem_i, double dt)
{
	double Ca_junc_observed;
	if(mem_i == 0) Ca_junc_observed = segments[seg_i].caUnits[0].Ca_srs;
	if(mem_i == 1) Ca_junc_observed = segments[seg_i].caUnits[settings->N_CaDomains-1].Ca_srs;
	
	double temp_IpCa, ECa_junc, ECa_sl;
	
	// Junctional SRS Component
	double ca_junc_pow = pow(Ca_junc_observed,1.6);
	temp_IpCa = data->IbarSLCaP * ca_junc_pow / (data->KmPCaPow1_6+ca_junc_pow);
	ECa_junc = 300;
	if(Ca_junc_observed > 1E-8)
	{
		ECa_junc = (1/data->frt/2) * log(settings->Ca_o/Ca_junc_observed);
	}
	segments[seg_i].membrane[mem_i].IpCa_junc = (1 - settings->IpCaB) * data->Fjunc * temp_IpCa;
	
	// Subsarcolemmal component
	double ca_sl_pow = pow(segments[seg_i].membrane[mem_i].Ca_sl,1.6);
	temp_IpCa = data->IbarSLCaP * ca_sl_pow / (data->KmPCaPow1_6+ca_sl_pow);
	ECa_sl = 300;
	if(segments[seg_i].membrane[mem_i].Ca_sl > 1E-8)
	{
		ECa_sl = (1/data->frt/2) * log(settings->Ca_o/segments[seg_i].membrane[mem_i].Ca_sl);
	}
	segments[seg_i].membrane[mem_i].IpCa_sl = (1 - settings->IpCaB) * (1 - data->Fjunc) * temp_IpCa;
		
	if(settings->Ca_o > 0)
	{
		segments[seg_i].membrane[mem_i].ICab_junc = (1 - settings->ICabB) * data->Fjunc * data->GCab * (Vm-ECa_junc);
		segments[seg_i].membrane[mem_i].ICab_sl = (1 - settings->ICabB) * (1 - data->Fjunc) * data->GCab * (Vm-ECa_sl);
	}
	else
	{
		segments[seg_i].membrane[mem_i].ICab_junc = 0;
		segments[seg_i].membrane[mem_i].ICab_sl = 0;
	}
	
	if(showlog)
	{
		cout << "  seg_i = " << seg_i << ", mem_i = " << mem_i << ", Ca_sl = " << segments[seg_i].membrane[mem_i].Ca_sl << ", temp_IpCa = " << temp_IpCa << ", ECa_sl = " << ECa_sl << ", ICab_sl = " << segments[seg_i].membrane[mem_i].ICab_sl << endl;
	}
}

void HAMM::updateNaDomain(double dt, double t, int segment, int ca_domain, double Na_i_L_Domain, double Na_i_R_Domain, double Na_i_U_Segment, double Na_i_D_Segment, double Na_srs_L_Domain, double Na_srs_R_Domain, double Na_srs_U_Segment, double Na_srs_D_Segment)
{
//	// Allow simulation at finer resolution
//	int numsteps = max(1, (int)(1 + dt / 0.075)); //settings->PARAM_dTmin);		
//	dt = dt / numsteps;
//	for(int abc=0; abc<numsteps; abc++)
//	{
		
	// Diffusive fluxes between neighboring CaRU's
	// ----------------------------------------------------
		double J_diff_i_subcell = (Na_i_L_Domain - segments[segment].caUnits[ca_domain].Na_i + Na_i_R_Domain - segments[segment].caUnits[ca_domain].Na_i) / data->tau_diff_Cai_Domain + (Na_i_U_Segment - segments[segment].caUnits[ca_domain].Na_i + Na_i_D_Segment - segments[segment].caUnits[ca_domain].Na_i) / data->tau_diff_Cai_Segment;
		double J_diff_srs_subcell = (Na_srs_L_Domain - segments[segment].caUnits[ca_domain].Na_srs + Na_srs_R_Domain - segments[segment].caUnits[ca_domain].Na_srs) / data->tau_diff_Casrs_Domain; 
		J_diff_srs_subcell += (Na_srs_U_Segment - segments[segment].caUnits[ca_domain].Na_srs + Na_srs_D_Segment - segments[segment].caUnits[ca_domain].Na_srs) / data->tau_diff_Casrs_Segment;
		
		double J_diff_SRS_i = (segments[segment].caUnits[ca_domain].Na_srs - segments[segment].caUnits[ca_domain].Na_i) / data->tau_diff_SRS_i;
		
	//Update local ion concentrations
	// ----------------------------------------------------
		// Sodium Buffering
		double dBuff_Na_junc = data->kon_na * segments[segment].caUnits[ca_domain].Na_srs * (data->Bmax_Naj-segments[segment].caUnits[ca_domain].Buff_Na_junc) - data->koff_na*segments[segment].caUnits[ca_domain].Buff_Na_junc; // NaBj      [mM/ms]
			
		double dNa_i, dNa_srs;
		dNa_srs = 0; dNa_i = 0;
		
		if(ca_domain == 0) 
		{
			//Leftmost unit, no release, but diffusion from membrane 0			
			//double I_Na_tot_sl = (2 * settings->N_Segments) * 3 * segments[segment].caUnits[ca_domain].INaCa_sl;
			dNa_i = data->J_ca_slmyo*0.5/(data->vmyo/settings->N_CaDomains)*(segments[segment].membrane[0].Na_sl - segments[segment].caUnits[ca_domain].Na_i) + J_diff_i_subcell + J_diff_SRS_i;
						
			double INa_tot_junc = segments[segment].membrane[0].INa_junc + segments[segment].membrane[0].INab_junc + (2 * settings->N_Segments) *3*segments[segment].caUnits[ca_domain].INaCa_junc + 3*segments[segment].membrane[0].INaK_junc;
			dNa_srs = -INa_tot_junc * data->CmemF/(data->vjunc*2) - dBuff_Na_junc + data->J_ca_juncsl*0.5/(data->vjunc/settings->N_CaDomains)*(segments[segment].membrane[0].Na_sl - segments[segment].caUnits[ca_domain].Na_srs) + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;
		}
		else
		{
			if(ca_domain == settings->N_CaDomains - 1)
			{
				//Rightmost unit, no release, but diffusion from membrane 1
				//double I_Na_tot_sl = (2 * settings->N_Segments) * 3 * segments[segment].caUnits[ca_domain].INaCa_sl;
				dNa_i = data->J_ca_slmyo*0.5/(data->vmyo/settings->N_CaDomains)*(segments[segment].membrane[1].Na_sl - segments[segment].caUnits[ca_domain].Na_i) + J_diff_i_subcell + J_diff_SRS_i;
						
				double INa_tot_junc = segments[segment].membrane[1].INa_junc + segments[segment].membrane[1].INab_junc + (2 * settings->N_Segments) *3*segments[segment].caUnits[ca_domain].INaCa_junc + 3*segments[segment].membrane[1].INaK_junc;
				dNa_srs = -INa_tot_junc * data->CmemF/(data->vjunc*2) - dBuff_Na_junc + data->J_ca_juncsl*0.5/(data->vjunc/settings->N_CaDomains)*(segments[segment].membrane[1].Na_sl - segments[segment].caUnits[ca_domain].Na_srs) + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;
			}
			else
			{
				//Other unit, release from corbular SR, no diffusion from membrane
				double I_Na_tot_i = (2 * settings->N_Segments) * 3 * segments[segment].caUnits[ca_domain].INaCa_sl;
				dNa_i = -I_Na_tot_i * data->CmemF/(data->vmyo*2) + J_diff_i_subcell + J_diff_SRS_i;
						
				double INa_tot_junc = (2 * settings->N_Segments) * 3 * segments[segment].caUnits[ca_domain].INaCa_junc;		// [uA/uF]
				dNa_srs = -INa_tot_junc * data->CmemF/(data->vjunc*2) - dBuff_Na_junc + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;
			}
		}
		
		//Update states
		segments[segment].caUnits[ca_domain].Na_i = check_State_Change(segments[segment].caUnits[ca_domain].Na_i, dNa_i, dt);	
		segments[segment].caUnits[ca_domain].Na_srs = check_State_Change(segments[segment].caUnits[ca_domain].Na_srs, dNa_srs, dt);					
		segments[segment].caUnits[ca_domain].Buff_Na_junc = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Na_junc, dBuff_Na_junc, dt);	
	//}
}

void HAMM::updateCaDomain(double dt, double t, int segment, int ca_domain, double Ca_i_L_Domain, double Ca_i_R_Domain, double Ca_i_U_Segment, double Ca_i_D_Segment, double Ca_sr_L_Domain, double Ca_sr_R_Domain, double Ca_sr_U_Segment, double Ca_sr_D_Segment, double Ca_srs_L_Domain, double Ca_srs_R_Domain, double Ca_srs_U_Segment, double Ca_srs_D_Segment)
{
	// Determine local ionic currents and fluxes
	// ----------------------------------------------------
	if(settings->RyRModelType == 2)	Ca_CICR(segment, ca_domain, dt, t);
	if(settings->RyRModelType == 3)	Ca_CICR_Stoch(segment, ca_domain, dt, t);
	
	// Allow simulation at finer resolution
	int numsteps = max(1, (int)(1 + dt / 0.075)); //settings->PARAM_dTmin);		
	dt = dt / numsteps;
	for(int abc=0; abc<numsteps; abc++)
	{
		
	// Diffusive fluxes between neighboring CaRU's
	// ----------------------------------------------------
		double J_diff_i_subcell = (Ca_i_L_Domain - segments[segment].caUnits[ca_domain].Ca_i + Ca_i_R_Domain - segments[segment].caUnits[ca_domain].Ca_i) / data->tau_diff_Cai_Domain + (Ca_i_U_Segment - segments[segment].caUnits[ca_domain].Ca_i + Ca_i_D_Segment - segments[segment].caUnits[ca_domain].Ca_i) / data->tau_diff_Cai_Segment;
		double J_diff_sr_subcell = (Ca_sr_L_Domain - segments[segment].caUnits[ca_domain].Ca_sr + Ca_sr_R_Domain - segments[segment].caUnits[ca_domain].Ca_sr) / data->tau_diff_Casr_Domain + (Ca_sr_U_Segment - segments[segment].caUnits[ca_domain].Ca_sr + Ca_sr_D_Segment - segments[segment].caUnits[ca_domain].Ca_sr) / data->tau_diff_Casr_Segment;
		
		double J_diff_srs_subcell = (Ca_srs_L_Domain - segments[segment].caUnits[ca_domain].Ca_srs + Ca_srs_R_Domain - segments[segment].caUnits[ca_domain].Ca_srs) / data->tau_diff_Casrs_Domain; 
		J_diff_srs_subcell += (Ca_srs_U_Segment - segments[segment].caUnits[ca_domain].Ca_srs + Ca_srs_D_Segment - segments[segment].caUnits[ca_domain].Ca_srs) / data->tau_diff_Casrs_Segment;
		
		double J_diff_SRS_i = (segments[segment].caUnits[ca_domain].Ca_srs - segments[segment].caUnits[ca_domain].Ca_i) / data->tau_diff_SRS_i;
			
	//Update local ion concentrations
	// ----------------------------------------------------
		// Cytosolic Ca Buffers
		double dBuff_Ca_TnCL = data->kon_tncl * segments[segment].caUnits[ca_domain].Ca_i * (settings->Ca_Buff_Factor * data->Bmax_TnClow - segments[segment].caUnits[ca_domain].Buff_Ca_TnCL) - data->koff_tncl * segments[segment].caUnits[ca_domain].Buff_Ca_TnCL;            		// TnCL      [mM/ms]
		double dBuff_Ca_TnCHc = data->kon_tnchca * segments[segment].caUnits[ca_domain].Ca_i * (settings->Ca_Buff_Factor * data->Bmax_TnChigh - segments[segment].caUnits[ca_domain].Buff_Ca_TnCHc - segments[segment].caUnits[ca_domain].Buff_Ca_TnCHm) - data->koff_tnchca * segments[segment].caUnits[ca_domain].Buff_Ca_TnCHc; 	// TnCHc     [mM/ms]
		double dBuff_Ca_TnCHm = data->kon_tnchmg * settings->Mg_i * (settings->Ca_Buff_Factor * data->Bmax_TnChigh - segments[segment].caUnits[ca_domain].Buff_Ca_TnCHc - segments[segment].caUnits[ca_domain].Buff_Ca_TnCHm) - data->koff_tnchmg * segments[segment].caUnits[ca_domain].Buff_Ca_TnCHm;   // TnCHm     [mM/ms]
		double dBuff_Ca_CaM = data->kon_cam * segments[segment].caUnits[ca_domain].Ca_i * (settings->Ca_Buff_Factor * data->Bmax_CaM - segments[segment].caUnits[ca_domain].Buff_Ca_CaM) - data->koff_cam * segments[segment].caUnits[ca_domain].Buff_Ca_CaM;                 // CaM       [mM/ms]
		double dBuff_Ca_Myosin_ca = data->kon_myoca * segments[segment].caUnits[ca_domain].Ca_i * (settings->Ca_Buff_Factor * data->Bmax_myosin - segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_ca - segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_mg) - data->koff_myoca * segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_ca;    // Myosin_ca [mM/ms]
		double dBuff_Ca_Myosin_mg = data->kon_myomg * settings->Mg_i * (settings->Ca_Buff_Factor * data->Bmax_myosin - segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_ca - segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_mg) - data->koff_myomg * segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_mg;      // Myosin_mg [mM/ms]
		double dBuff_Ca_SRB = data->kon_sr * segments[segment].caUnits[ca_domain].Ca_i * (settings->Ca_Buff_Factor * data->Bmax_SR - segments[segment].caUnits[ca_domain].Buff_Ca_SRB) - data->koff_sr * segments[segment].caUnits[ca_domain].Buff_Ca_SRB;                    // SRB       [mM/ms]
		double dBuff_Ca_EGTA_cyt = (0.005 / dt) * (data->kon_EGTA * segments[segment].caUnits[ca_domain].Ca_i * (data->EGTA - segments[segment].caUnits[ca_domain].Buff_Ca_EGTA_cyt) - data->koff_EGTA * segments[segment].caUnits[ca_domain].Buff_Ca_EGTA_cyt);
		double J_CaB_cytosol = dBuff_Ca_TnCL + dBuff_Ca_TnCHc + dBuff_Ca_TnCHm + dBuff_Ca_CaM + dBuff_Ca_Myosin_ca + dBuff_Ca_Myosin_mg + dBuff_Ca_SRB + dBuff_Ca_EGTA_cyt;
			
		double dCa_i, dCa_srs;
		dCa_srs = 0; dCa_i = 0;
		
		if(ca_domain == 0) 
		{
			//Leftmost unit, no release, but diffusion from membrane 0
			dCa_i = -segments[segment].caUnits[ca_domain].Jup * data->vsr/data->vmyo - J_CaB_cytosol + data->J_ca_slmyo*0.5/(data->vmyo/settings->N_CaDomains)*(segments[segment].membrane[0].Ca_sl - segments[segment].caUnits[ca_domain].Ca_i) + J_diff_i_subcell + J_diff_SRS_i;
			
			//double I_Ca_tot_junc = segments[segment].membrane[0].ICaL_junc + segments[segment].membrane[0].ICab_junc + segments[segment].membrane[0].IpCa_junc - 2*segments[segment].membrane[0].INaCa_junc;		// [uA/uF]
			double I_Ca_tot_junc = (2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].ICaL_junc + segments[segment].membrane[0].ICab_junc + segments[segment].membrane[0].IpCa_junc - 2*(2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].INaCa_junc;		// [uA/uF]
			dCa_srs = -I_Ca_tot_junc * data->CmemF/(data->vjunc*2) + data->J_ca_juncsl*0.5/(data->vjunc/settings->N_CaDomains)*(segments[segment].membrane[0].Ca_sl - segments[segment].caUnits[ca_domain].Ca_srs) + segments[segment].caUnits[ca_domain].Jrel*data->vsr/data->vjunc + segments[segment].caUnits[ca_domain].Jleak*data->vmyo/data->vjunc + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;
		}
		else
		{
			if(ca_domain == settings->N_CaDomains - 1)
			{
				//Rightmost unit, no release, but diffusion from membrane 1
				dCa_i = -segments[segment].caUnits[ca_domain].Jup * data->vsr/data->vmyo - J_CaB_cytosol + data->J_ca_slmyo*0.5/(data->vmyo/settings->N_CaDomains)*(segments[segment].membrane[1].Ca_sl - segments[segment].caUnits[ca_domain].Ca_i) + J_diff_i_subcell + J_diff_SRS_i;

				//double I_Ca_tot_junc = segments[segment].membrane[1].ICaL_junc + segments[segment].membrane[1].ICab_junc + segments[segment].membrane[1].IpCa_junc - 2*segments[segment].membrane[1].INaCa_junc;		// [uA/uF]
				double I_Ca_tot_junc = (2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].ICaL_junc + segments[segment].membrane[1].ICab_junc + segments[segment].membrane[1].IpCa_junc - 2*(2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].INaCa_junc;		// [uA/uF]
				dCa_srs = -I_Ca_tot_junc * data->CmemF/(data->vjunc*2) + data->J_ca_juncsl*0.5/(data->vjunc/settings->N_CaDomains)*(segments[segment].membrane[1].Ca_sl - segments[segment].caUnits[ca_domain].Ca_srs) + segments[segment].caUnits[ca_domain].Jrel*data->vsr/data->vjunc + segments[segment].caUnits[ca_domain].Jleak*data->vmyo/data->vjunc + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;
			}
			else
			{
				double I_Ca_tot_i = - 2*(2 * settings->N_Segments)*segments[segment].caUnits[ca_domain].INaCa_sl; //I_Ca_tot_i = 0;
				dCa_i = -I_Ca_tot_i * data->CmemF/(data->vmyo*2) - segments[segment].caUnits[ca_domain].Jup * data->vsr/data->vmyo - J_CaB_cytosol + J_diff_i_subcell + J_diff_SRS_i;
				
				double I_Ca_tot_junc = (2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].ICaL_junc - 2*(2 * settings->N_Segments) * segments[segment].caUnits[ca_domain].INaCa_junc;	// [uA/uF]
				dCa_srs = -I_Ca_tot_junc * data->CmemF/(data->vjunc*2) + segments[segment].caUnits[ca_domain].Jrel*data->vsr/data->vjunc + segments[segment].caUnits[ca_domain].Jleak*data->vmyo/data->vjunc + J_diff_srs_subcell - data->vmyo/data->vjunc * J_diff_SRS_i;				
			}
		}
		
		//Update states
		double cor_fact = 1.0;
		segments[segment].caUnits[ca_domain].Ca_i = check_State_Change(segments[segment].caUnits[ca_domain].Ca_i, dCa_i, dt);	
				
		segments[segment].caUnits[ca_domain].J_diff_i_subcell = J_diff_i_subcell;
		
		double Km_SLL = data->koff_sll / data->kon_sll;
		double Km_SLH = data->koff_slh / data->kon_slh;
			
		double Ca_srs_tot = segments[segment].caUnits[ca_domain].Ca_srs + segments[segment].caUnits[ca_domain].Buff_Ca_SLLsrs + segments[segment].caUnits[ca_domain].Buff_Ca_SLHsrs;
		Ca_srs_tot = check_State_Change(Ca_srs_tot, dCa_srs, dt);
		segments[segment].caUnits[ca_domain].Ca_srs = Buff_Ca_DoubleBuff(Ca_srs_tot, data->Bmax_SLlowj, data->Bmax_SLhighj, Km_SLL, Km_SLH);
		segments[segment].caUnits[ca_domain].Buff_Ca_SLLsrs = segments[segment].caUnits[ca_domain].Ca_srs * data->Bmax_SLlowj / (segments[segment].caUnits[ca_domain].Ca_srs + Km_SLL);
		segments[segment].caUnits[ca_domain].Buff_Ca_SLHsrs = Ca_srs_tot - segments[segment].caUnits[ca_domain].Ca_srs - segments[segment].caUnits[ca_domain].Buff_Ca_SLLsrs;
		
		double Ca_sr_tot = segments[segment].caUnits[ca_domain].Ca_sr + segments[segment].caUnits[ca_domain].Buff_Ca_CSQN;
		double dCa_sr_tot = segments[segment].caUnits[ca_domain].Jup - (segments[segment].caUnits[ca_domain].Jleak * data->vmyo/data->vsr + segments[segment].caUnits[ca_domain].Jrel) + J_diff_sr_subcell;         // Ca_sr     [mM/ms] %Ratio 3 leak current
		Ca_sr_tot = check_State_Change(Ca_sr_tot, dCa_sr_tot, dt);
		double b_Casr_buff = Ca_sr_tot + data->Bmax_Csqn + data->koff_csqn / data->kon_csqn;
		double c_Casr_buff = Ca_sr_tot * data->Bmax_Csqn;
		segments[segment].caUnits[ca_domain].Buff_Ca_CSQN = 0.5 * (b_Casr_buff - sqrt(b_Casr_buff*b_Casr_buff - 4*c_Casr_buff));
		segments[segment].caUnits[ca_domain].Ca_sr = Ca_sr_tot - segments[segment].caUnits[ca_domain].Buff_Ca_CSQN;
				
		segments[segment].caUnits[ca_domain].Buff_Ca_TnCL = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_TnCL, dBuff_Ca_TnCL*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_TnCHc = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_TnCHc, dBuff_Ca_TnCHc*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_TnCHm = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_TnCHm, dBuff_Ca_TnCHm*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_CaM = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_CaM, dBuff_Ca_CaM*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_ca = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_ca, dBuff_Ca_Myosin_ca*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_mg = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_Myosin_mg, dBuff_Ca_Myosin_mg*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_SRB = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_SRB, dBuff_Ca_SRB*cor_fact, dt);
		segments[segment].caUnits[ca_domain].Buff_Ca_EGTA_cyt = check_State_Change(segments[segment].caUnits[ca_domain].Buff_Ca_EGTA_cyt, dBuff_Ca_EGTA_cyt*cor_fact, dt);
	}
}

void HAMM::UpdateConcentrations(int seg_i, int mem_i, double dt, double t, double Na_sl_U_Segment, double Na_sl_D_Segment, double Ca_sl_U_Segment, double Ca_sl_D_Segment, double Ca_i_neighbor, double Ca_srs_neighbor, double Na_i_neighbor)
{
	// Sodium Buffering and Concentrations
	double dBuff_Na_sl = data->kon_na * segments[seg_i].membrane[mem_i].Na_sl * (data->Bmax_Nasl-segments[seg_i].membrane[mem_i].Buff_Na_sl) - data->koff_na*segments[seg_i].membrane[mem_i].Buff_Na_sl;        // NaBsl     [mM/ms]
	double J_diff_sl_subcell = (Na_sl_U_Segment - segments[seg_i].membrane[mem_i].Na_sl + Na_sl_D_Segment - segments[seg_i].membrane[mem_i].Na_sl) / data->tau_diff_Casl_Segment;

	double dNa_sl, INa_tot_sl;
	if(mem_i == 0)
	{
		INa_tot_sl = segments[seg_i].membrane[mem_i].INa_sl + segments[seg_i].membrane[mem_i].INab_sl + 3*segments[seg_i].caUnits[0].INaCa_sl + 3*segments[seg_i].membrane[mem_i].INaK_sl + segments[seg_i].caUnits[0].ICaL_Na_sl + segments[seg_i].membrane[mem_i].INaL_sl;	
		dNa_sl = -INa_tot_sl * data->CmemF/data->vsl + data->J_na_juncsl/data->vsl * (segments[seg_i].caUnits[0].Na_srs - segments[seg_i].membrane[mem_i].Na_sl) + data->J_na_slmyo/(0.5 * data->vsl) * (Na_i_neighbor - segments[seg_i].membrane[mem_i].Na_sl) - dBuff_Na_sl + J_diff_sl_subcell;
	}
	else
	{
		INa_tot_sl = segments[seg_i].membrane[mem_i].INa_sl + segments[seg_i].membrane[mem_i].INab_sl + 3*segments[seg_i].caUnits[settings->N_CaDomains-1].INaCa_sl + 3*segments[seg_i].membrane[mem_i].INaK_sl + segments[seg_i].caUnits[settings->N_CaDomains-1].ICaL_Na_sl + segments[seg_i].membrane[mem_i].INaL_sl;	
		dNa_sl = -INa_tot_sl * data->CmemF/data->vsl + data->J_na_juncsl/data->vsl * (segments[seg_i].caUnits[settings->N_CaDomains-1].Na_srs - segments[seg_i].membrane[mem_i].Na_sl) + data->J_na_slmyo/(0.5 * data->vsl) * (Na_i_neighbor - segments[seg_i].membrane[mem_i].Na_sl) - dBuff_Na_sl + J_diff_sl_subcell;
	}

	segments[seg_i].membrane[mem_i].Buff_Na_sl = check_State_Change(segments[seg_i].membrane[mem_i].Buff_Na_sl, dBuff_Na_sl, dt);
	segments[seg_i].membrane[mem_i].Na_sl = check_State_Change(segments[seg_i].membrane[mem_i].Na_sl, dNa_sl, dt);

	double I_Ca_tot_sl;
	if(mem_i == 0) I_Ca_tot_sl = segments[seg_i].membrane[mem_i].ICab_sl + segments[seg_i].membrane[mem_i].IpCa_sl - 2*(2*settings->N_Segments)*segments[seg_i].caUnits[0].INaCa_sl;            		// [uA/uF]
	if(mem_i == 1) I_Ca_tot_sl = segments[seg_i].membrane[mem_i].ICab_sl + segments[seg_i].membrane[mem_i].IpCa_sl - 2*(2*settings->N_Segments)*segments[seg_i].caUnits[settings->N_CaDomains-1].INaCa_sl;            		// [uA/uF]
	double J_diff_Casl_subcell = (Ca_sl_U_Segment - segments[seg_i].membrane[mem_i].Ca_sl + Ca_sl_D_Segment - segments[seg_i].membrane[mem_i].Ca_sl) / data->tau_diff_Casl_Segment;

	double dBuff_Ca_EGTA_sl = (0.005 / dt) * (data->kon_EGTA * segments[seg_i].membrane[mem_i].Ca_sl * (data->EGTA * (data->vmyo / data->vsl) - segments[seg_i].membrane[mem_i].Buff_Ca_EGTA_sl) - data->koff_EGTA * segments[seg_i].membrane[mem_i].Buff_Ca_EGTA_sl);
	double Km_SLL = data->koff_sll / data->kon_sll;
	double Km_SLH = data->koff_slh / data->kon_slh;
		
	double Ca_junc_tot = segments[seg_i].membrane[mem_i].Ca_junc + segments[seg_i].membrane[mem_i].Buff_Ca_SLLj + segments[seg_i].membrane[mem_i].Buff_Ca_SLHj;
	double Ca_sl_tot = segments[seg_i].membrane[mem_i].Ca_sl + segments[seg_i].membrane[mem_i].Buff_Ca_SLLsl + segments[seg_i].membrane[mem_i].Buff_Ca_SLHsl;
	
	double dCa_junc_tot, dCa_sl_tot;
	if(mem_i == 0) dCa_junc_tot = data->J_ca_juncsl/data->vjunc*(segments[seg_i].membrane[mem_i].Ca_sl - segments[seg_i].membrane[mem_i].Ca_junc) + (segments[seg_i].caUnits[0].Ca_srs - segments[seg_i].membrane[mem_i].Ca_junc) / data->tau_diff_Casrs_Domain;  // Ca_j
	if(mem_i == 1) dCa_junc_tot = data->J_ca_juncsl/data->vjunc*(segments[seg_i].membrane[mem_i].Ca_sl - segments[seg_i].membrane[mem_i].Ca_junc) + (segments[seg_i].caUnits[settings->N_CaDomains-1].Ca_srs - segments[seg_i].membrane[mem_i].Ca_junc) / data->tau_diff_Casrs_Domain;  // Ca_j		
	dCa_sl_tot = -I_Ca_tot_sl * data->CmemF/(data->vsl*2) + data->J_ca_juncsl/data->vsl*(segments[seg_i].membrane[mem_i].Ca_junc - segments[seg_i].membrane[mem_i].Ca_sl) + data->J_ca_slmyo/(data->vsl)*(Ca_i_neighbor - segments[seg_i].membrane[mem_i].Ca_sl) + J_diff_Casl_subcell - dBuff_Ca_EGTA_sl;   // Ca_sl
		
	dCa_junc_tot = 0;			
	dCa_sl_tot = -I_Ca_tot_sl * data->CmemF/(data->vsl*2) + data->J_ca_juncsl/data->vsl*(Ca_srs_neighbor - segments[seg_i].membrane[mem_i].Ca_sl) + data->J_ca_slmyo/(data->vsl)*(Ca_i_neighbor - segments[seg_i].membrane[mem_i].Ca_sl) + J_diff_Casl_subcell - dBuff_Ca_EGTA_sl;   // Ca_sl						

	Ca_junc_tot = check_State_Change(Ca_junc_tot, dCa_junc_tot, dt);
	segments[seg_i].membrane[mem_i].Ca_junc = Buff_Ca_DoubleBuff(Ca_junc_tot, data->Bmax_SLlowj, data->Bmax_SLhighj, Km_SLL, Km_SLH);
	segments[seg_i].membrane[mem_i].Buff_Ca_SLLj = segments[seg_i].membrane[mem_i].Ca_junc * data->Bmax_SLlowj / (segments[seg_i].membrane[mem_i].Ca_junc + Km_SLL);
	segments[seg_i].membrane[mem_i].Buff_Ca_SLHj = Ca_junc_tot - segments[seg_i].membrane[mem_i].Ca_junc - segments[seg_i].membrane[mem_i].Buff_Ca_SLLj;
				
	Ca_sl_tot = check_State_Change(Ca_sl_tot, dCa_sl_tot, dt);
	segments[seg_i].membrane[mem_i].Ca_sl = Buff_Ca_DoubleBuff(Ca_sl_tot, data->Bmax_SLlowsl, data->Bmax_SLhighsl, Km_SLL, Km_SLH);
	segments[seg_i].membrane[mem_i].Buff_Ca_SLLsl = segments[seg_i].membrane[mem_i].Ca_sl * data->Bmax_SLlowsl / (segments[seg_i].membrane[mem_i].Ca_sl + Km_SLL);
	segments[seg_i].membrane[mem_i].Buff_Ca_SLHsl = Ca_sl_tot - segments[seg_i].membrane[mem_i].Ca_sl - segments[seg_i].membrane[mem_i].Buff_Ca_SLLsl;
	segments[seg_i].membrane[mem_i].Buff_Ca_EGTA_sl = check_State_Change(segments[seg_i].membrane[mem_i].Buff_Ca_EGTA_sl, dBuff_Ca_EGTA_sl, dt);
}

//Stochastic Ryanodine Receptor formulation
void HAMM::Ca_CICR_Stoch(int seg_i, int cadom_i, double dt, double t)
{
	if(settings->RyR_Distribution_Scaling[seg_i][cadom_i] > 1E-12)	
	{
		double kCaSR;
		
		double Ca_cis_observed = segments[seg_i].caUnits[cadom_i].Ca_srs;
		double Ca_sr_observed = segments[seg_i].caUnits[cadom_i].Ca_sr;
		
		segments[seg_i].caUnits[cadom_i].Jrel = (1 - settings->IrelB) * (settings->RyR_Distribution_Scaling[seg_i][cadom_i] * settings->N_Segments * settings->N_CaDomains) * settings->RyRParams_NP[seg_i][cadom_i][11] * settings->N_CaDomains * segments[seg_i].caUnits[cadom_i].RyRState_NP[1] / ((1.0 * settings->numRyRs) / (settings->N_Segments * settings->N_CaDomains)) * (Ca_sr_observed-Ca_cis_observed);          // [mM/ms]
		segments[seg_i].caUnits[cadom_i].Jleak = 0; //(1 - settings->IrelB) * Jleakmax * (segments[seg_i].caUnits[cadom_i].Ca_sr-Ca_cis_observed);           //   [mM/ms]	
		
		if (settings->RyRParams_NP[seg_i][cadom_i][21] > 0 && settings->RyRParams_NP[seg_i][cadom_i][21] < 100 && Ca_sr_observed > settings->RyRParams_NP[seg_i][cadom_i][21]) Ca_sr_observed = settings->RyRParams_NP[seg_i][cadom_i][21];
		
		// RyR Gating
		double exp_Ca_cis, Act_ss, tau_Act, Act_ss_low;
		
		double deltaf = 1; // + 2 * (settings->RyRParams_NP[19] - 1) / (1 + pow(Ca_sr_observed / settings->RyRParams_NP[1], settings->RyRParams_NP[2]));	
		exp_Ca_cis = 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][4]*deltaf)/settings->RyRParams_NP[seg_i][cadom_i][5]));
		Act_ss = (1 / (1 + pow(segments[seg_i].caUnits[cadom_i].RyR_Integr / 0.5, 8))) * (settings->RyRParams_NP[seg_i][cadom_i][6] + settings->RyRParams_NP[seg_i][cadom_i][3] * exp_Ca_cis);	
		Act_ss_low = (1 / (1 + pow(segments[seg_i].caUnits[cadom_i].RyR_Integr / 0.5, 8))) * (settings->RyRParams_NP[seg_i][cadom_i][6] + settings->RyRParams_NP[seg_i][cadom_i][3] * (1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][20] * settings->RyRParams_NP[seg_i][cadom_i][4]*deltaf)/settings->RyRParams_NP[seg_i][cadom_i][5]))));	
		tau_Act = settings->RyRParams_NP[seg_i][cadom_i][7] + settings->RyRParams_NP[seg_i][cadom_i][8] * (1 - 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][18])/settings->RyRParams_NP[seg_i][cadom_i][5])));
		double tau_Act_low = tau_Act; //settings->RyRParams_NP[seg_i][cadom_i][25] * tau_Act;	
		double midlev = settings->RyRParams_NP[seg_i][cadom_i][0]; // + (1 - settings->RyRParams_NP[0]) / (1 + pow(0.15 / segments[seg_i].caUnits[cadom_i].RyR_Integr, 8));
		double Inact_ss = midlev / (1 + pow(Ca_sr_observed / 4,4)) + (settings->RyRParams_NP[seg_i][cadom_i][9] - midlev) / (1 + pow(Ca_sr_observed / settings->RyRParams_NP[seg_i][cadom_i][1], settings->RyRParams_NP[seg_i][cadom_i][2]));
		double tau_Inact = settings->RyRParams_NP[seg_i][cadom_i][10] + settings->RyRParams_NP[seg_i][cadom_i][12] / (1 + pow(settings->RyRParams_NP[seg_i][cadom_i][1] / Ca_sr_observed, settings->RyRParams_NP[seg_i][cadom_i][2]));
		
		double tau_Flec = 100;
		double Flec_Block_ss = 0;
		if(settings->Flec > 0)
		{
			tau_Flec = settings->RyRParams_NP[seg_i][cadom_i][24];
			Flec_Block_ss = 1.0 / (1 + pow(settings->RyRParams_NP[seg_i][cadom_i][22] / settings->Flec, settings->RyRParams_NP[seg_i][cadom_i][23]));
		}
		
		//Methodology to clamp RyR2
		if(settings->clamp_RyR_Po >= -0.1 && settings->clamp_RyR_Po <= 1) 
		{
			if(settings->numRyR_Po_Indices < 0 || GenPurpose::occursInList(settings->RyR_Po_Indices, settings->numRyR_Po_Indices, (seg_i - 1) * settings->N_CaDomains + cadom_i))
			{
				if(settings->clamp_RyR_Po >= 0)
				{
					Act_ss = settings->clamp_RyR_Po;
				
					if(settings->clamp_RyR_Po >= 0.3) 
					{
						Inact_ss = 0;
						Flec_Block_ss = 0;
						segments[seg_i].caUnits[cadom_i].Jrel = segments[seg_i].caUnits[cadom_i].Jrel * 2;
					}	
					exp_Ca_cis = 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][4])/settings->RyRParams_NP[seg_i][cadom_i][5]));
					tau_Act = settings->RyRParams_NP[seg_i][cadom_i][7] + settings->RyRParams_NP[seg_i][cadom_i][8] * kCaSR * (1 - exp_Ca_cis);
				}
				else
				{
					Act_ss = (-10 * settings->clamp_RyR_Po) * Act_ss;
				}
			}
		}
		
		double k_12 = Act_ss / tau_Act;
		double k_21 = (1 - Act_ss) / tau_Act;
		double k_43 = Act_ss_low / tau_Act_low;
		double k_34 = (1 - Act_ss_low) / tau_Act_low;
		double k_23 = Inact_ss / tau_Inact;
		double k_32 = (1 - Inact_ss) / tau_Inact;
		double k_41 = 0.1 * k_32;
		double k_14 = (k_12 * k_23 * k_34 * k_41) / (k_43 * k_32 * k_21); 
		
		double k_26 = Flec_Block_ss / tau_Flec;
		double k_62 = (1 - Flec_Block_ss) / tau_Flec;
		
		// Used to ensure stepsize adaptation
		double RyRStateSum = segments[seg_i].caUnits[cadom_i].RyRState_NP[0] + segments[seg_i].caUnits[cadom_i].RyRState_NP[1] + segments[seg_i].caUnits[cadom_i].RyRState_NP[2] + segments[seg_i].caUnits[cadom_i].RyRState_NP[3];
		double dR = ((k_41*segments[seg_i].caUnits[cadom_i].RyRState_NP[3]-k_14*segments[seg_i].caUnits[cadom_i].RyRState_NP[0])-(k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[0]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[1])) / RyRStateSum;   // R
		double dO = ((k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[0]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[1])-(k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[1]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[2])) / RyRStateSum; // O
		double dI = ((k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[1]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[2])-(k_34*segments[seg_i].caUnits[cadom_i].RyRState_NP[2]-k_43*segments[seg_i].caUnits[cadom_i].RyRState_NP[3])) / RyRStateSum;   // I
			check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[0] / RyRStateSum, dR, dt);	
			check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[1] / RyRStateSum, dO, dt);
			check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[2] / RyRStateSum, dI, dt);
		
			// Build probability matrix
			// For methodology see Heijman et al. PLoS Comput Biol 2013.
			double RyR_Matrix[8][8];
			double dt_used = dt;
				RyR_Matrix[0][0] = 1 - dt_used * (k_12 + k_14); RyR_Matrix[0][1] = dt_used * k_12; RyR_Matrix[0][2] = 0; RyR_Matrix[0][3] = dt_used * k_14; RyR_Matrix[0][4] = 0; RyR_Matrix[0][5] = 0; RyR_Matrix[0][6] = 0; RyR_Matrix[0][7] = 0;
				RyR_Matrix[1][0] = dt_used * k_21; RyR_Matrix[1][1] = 1 - dt_used * (k_21+k_23+k_26); RyR_Matrix[1][2] = dt_used * k_23; RyR_Matrix[1][3] = 0; RyR_Matrix[1][4] = 0; RyR_Matrix[1][5] = dt_used * k_26; RyR_Matrix[1][6] = 0; RyR_Matrix[1][7] = 0;
				RyR_Matrix[2][0] = 0; RyR_Matrix[2][1] = dt_used * k_32; RyR_Matrix[2][2] = 1 - dt_used * (k_32 + k_34); RyR_Matrix[2][3] = dt_used * k_34; RyR_Matrix[2][4] = 0; RyR_Matrix[2][5] = 0; RyR_Matrix[2][6] = 0; RyR_Matrix[2][7] = 0;
				RyR_Matrix[3][0] = dt_used * k_41; RyR_Matrix[3][1] = 0; RyR_Matrix[3][2] = dt_used * k_43; RyR_Matrix[3][3] = 1 - dt_used * (k_41 + k_43); RyR_Matrix[3][4] = 0; RyR_Matrix[3][5] = 0; RyR_Matrix[3][6] = 0; RyR_Matrix[3][7] = 0;
			
				RyR_Matrix[4][4] = 1 - dt_used * (k_12 + k_14); RyR_Matrix[4][5] = dt_used * k_12; RyR_Matrix[4][6] = 0; RyR_Matrix[4][7] = dt_used * k_14; RyR_Matrix[4][0] = 0; RyR_Matrix[4][1] = 0; RyR_Matrix[4][2] = 0; RyR_Matrix[4][3] = 0;
				RyR_Matrix[5][4] = dt_used * k_21; RyR_Matrix[5][5] = 1 - dt_used * (k_21+k_23+k_62); RyR_Matrix[5][6] = dt_used * k_23; RyR_Matrix[5][7] = 0; RyR_Matrix[5][0] = 0; RyR_Matrix[5][1] = dt_used * k_62; RyR_Matrix[5][2] = 0; RyR_Matrix[5][3] = 0;
				RyR_Matrix[6][4] = 0; RyR_Matrix[6][5] = dt_used * k_32; RyR_Matrix[6][6] = 1 - dt_used * (k_32 + k_34); RyR_Matrix[6][7] = dt_used * k_34; RyR_Matrix[6][0] = 0; RyR_Matrix[6][1] = 0; RyR_Matrix[6][2] = 0; RyR_Matrix[6][3] = 0;
				RyR_Matrix[7][4] = dt_used * k_41; RyR_Matrix[7][5] = 0; RyR_Matrix[7][6] = dt_used * k_43; RyR_Matrix[7][7] = 1 - dt_used * (k_41 + k_43); RyR_Matrix[7][0] = 0; RyR_Matrix[7][1] = 0; RyR_Matrix[7][2] = 0; RyR_Matrix[7][3] = 0;			
			
			
			// Compute state transitions		
				const int nrOfStatesInModel = 8;
				int newRyRState[nrOfStatesInModel];		
				for(int i=0; i<nrOfStatesInModel; i++) { newRyRState[i] = 0;	}
						
				for(int i=0; i<nrOfStatesInModel; i++)
				{
					if(segments[seg_i].caUnits[cadom_i].RyRState_NP[i] >= 1 && segments[seg_i].caUnits[cadom_i].RyRState_NP[i] <= 10) //Use uniform dist approach
					{
						for(int k=0; k<segments[seg_i].caUnits[cadom_i].RyRState_NP[i]; k++)
						{
							double cumulprob = 0.0;
							double rv = rvgen.rand();
							
							for(int j=0; j<nrOfStatesInModel; j++)
							{
								cumulprob += RyR_Matrix[i][j];
								if(cumulprob > rv)
								{
									newRyRState[j]++;
									break;
								}
							}
						}
					}
					else
					{
						if(segments[seg_i].caUnits[cadom_i].RyRState_NP[i] > 10) // Use binomial approach				
						{
							int numchannels = (int)segments[seg_i].caUnits[cadom_i].RyRState_NP[i];
							for(int dd=0; dd<nrOfStatesInModel; dd++)
							{
								if(dd != i && RyR_Matrix[i][dd] > 0) //Only take states other than self with nonzero chance
								{
									int numchannelsout = GenPurpose::applyBinomialInverseTransform(&rvgen, (int)segments[seg_i].caUnits[cadom_i].RyRState_NP[i], RyR_Matrix[i][dd]);
									if(numchannelsout > numchannels) numchannelsout = numchannels;
									newRyRState[dd] += numchannelsout;
									numchannels -= numchannelsout;
								}
							}
							newRyRState[i] += numchannels; // The remaining channels stay where they are
						}
					}		
				}

				//Ability to set channel state to a specific configuration
				if(settings->set_RyR_State != NULL) 
				{
					int dom_index = (seg_i * settings->N_CaDomains + cadom_i);
					double* curDomRyRState = settings->set_RyR_State[dom_index];
					
					if(curDomRyRState[0] >= 0)
					{
						newRyRState[0] = curDomRyRState[0];
						newRyRState[1] = curDomRyRState[1];
						newRyRState[2] = curDomRyRState[2];
						newRyRState[3] = curDomRyRState[3];
					}
				}
				
				for(int i=0; i<nrOfStatesInModel; i++) { segments[seg_i].caUnits[cadom_i].RyRState_NP[i] = newRyRState[i];	}
			
		
		double RyR_SEP_SS = 1.0 / (1.0 + exp((Ca_cis_observed - 1.8E-3) / 1E-6)); //1 / (1 + pow(Ca_cis_observed / 2.75E-4, 20));
		double tau_RyR_SEP = 60;
		segments[seg_i].caUnits[cadom_i].RyR_SEP = check_State_Change(segments[seg_i].caUnits[cadom_i].RyR_SEP, (RyR_SEP_SS - segments[seg_i].caUnits[cadom_i].RyR_SEP) / tau_RyR_SEP, dt);

		double dRyR_Int = 1 * segments[seg_i].caUnits[cadom_i].Jrel - 0.1 / (1 + exp((segments[seg_i].caUnits[cadom_i].Jrel - 1.5E-3) / 1E-4)) * segments[seg_i].caUnits[cadom_i].RyR_Integr;
		segments[seg_i].caUnits[cadom_i].RyR_Integr = check_State_Change(segments[seg_i].caUnits[cadom_i].RyR_Integr, dRyR_Int, dt);
	}
	else
	{
		segments[seg_i].caUnits[cadom_i].Jrel = 0.0;     // [mM/ms]
		segments[seg_i].caUnits[cadom_i].Jleak = 0; //(1 - settings->IrelB) * Jleakmax * (segments[seg_i].caUnits[cadom_i].Ca_sr-Ca_cis_observed);           //   [mM/ms]		
	}
	
	if(settings->SERCA_Distribution_Scaling[seg_i][cadom_i] > 1E-12)
	{
		// SERCA
		double tval_Kmr = pow(segments[seg_i].caUnits[cadom_i].Ca_sr / data->Kmr, data->hillSRCaP);
		double tval_Kmf_NP = pow(segments[seg_i].caUnits[cadom_i].Ca_i / data->Kmf_NP, data->hillSRCaP);
		double Jup_NP = (1 - settings->IupB) * (settings->SERCA_Distribution_Scaling[seg_i][cadom_i] * settings->N_Segments * settings->N_CaDomains) * data->Vmax_SRCaP * (tval_Kmf_NP - tval_Kmr) / (1 + tval_Kmf_NP + tval_Kmr);
		segments[seg_i].caUnits[cadom_i].Jup = Jup_NP;
	}
}

// Deterministic RyR2 simulation
void HAMM::Ca_CICR(int seg_i, int cadom_i, double dt, double t)
{
	if(settings->RyR_Distribution_Scaling[seg_i][cadom_i] > 1E-12)
	{
		double kCaSR;
			
		double Ca_cis_observed = segments[seg_i].caUnits[cadom_i].Ca_srs;
		if(segments[seg_i].caUnits[cadom_i].RyRState_NP[1] > 2 || segments[seg_i].caUnits[cadom_i].RyRState_NP[0] > 2)
		{
			cout << "Stochastic RyR state being used in deterministic simulations!! - t = " << t << ", dt = " << dt << ", seg_i = " << seg_i << ", cadom_i = " << cadom_i << endl;
			GenPurpose::wait(2.0);
		}
		
		segments[seg_i].caUnits[cadom_i].Jrel = (1 - settings->IrelB) * (settings->RyR_Distribution_Scaling[seg_i][cadom_i] * settings->N_Segments * settings->N_CaDomains) * settings->RyRParams_NP[seg_i][cadom_i][11] * settings->N_CaDomains * segments[seg_i].caUnits[cadom_i].RyRState_NP[1] * (segments[seg_i].caUnits[cadom_i].Ca_sr-Ca_cis_observed);          // [mM/ms]
		segments[seg_i].caUnits[cadom_i].Jleak = 0; //(1 - settings->IrelB) * Jleakmax * (segments[seg_i].caUnits[cadom_i].Ca_sr-Ca_cis_observed);           //   [mM/ms]

		double k_12, k_21, k_14, k_41;
		double exp_Ca_cis, Act_ss, tau_Act, Act_ss_low;
		double Inact_ss, tau_Inact;

		double Ca_sr_observed = segments[seg_i].caUnits[cadom_i].Ca_sr;
		if (settings->RyRParams_NP[seg_i][cadom_i][21] > 0 && settings->RyRParams_NP[seg_i][cadom_i][21] < 100 && Ca_sr_observed > settings->RyRParams_NP[seg_i][cadom_i][21]) Ca_sr_observed = settings->RyRParams_NP[seg_i][cadom_i][21];
			
		double deltaf = 1; //1 + 2 * (settings->RyRParams_NP[19] - 1) / (1 + pow(Ca_sr_observed / settings->RyRParams_NP[1], settings->RyRParams_NP[2]));
		exp_Ca_cis = 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][4]*deltaf)/settings->RyRParams_NP[seg_i][cadom_i][5]));
		
		Act_ss =  (1 / (1 + pow(segments[seg_i].caUnits[cadom_i].RyR_Integr / 0.5, 8))) * (settings->RyRParams_NP[seg_i][cadom_i][6] + settings->RyRParams_NP[seg_i][cadom_i][3] * exp_Ca_cis);		
		Act_ss_low =  (1 / (1 + pow(segments[seg_i].caUnits[cadom_i].RyR_Integr / 0.5, 8))) * (settings->RyRParams_NP[seg_i][cadom_i][6] + settings->RyRParams_NP[seg_i][cadom_i][3] * (1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][20] * settings->RyRParams_NP[seg_i][cadom_i][4]*deltaf)/settings->RyRParams_NP[seg_i][cadom_i][5]))));
		tau_Act = settings->RyRParams_NP[seg_i][cadom_i][7] + settings->RyRParams_NP[seg_i][cadom_i][8] * (1 - 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][18])/settings->RyRParams_NP[seg_i][cadom_i][5])));
		double tau_Act_low = tau_Act; //settings->RyRParams_NP[seg_i][cadom_i][25] * tau_Act;
		double midlev = settings->RyRParams_NP[seg_i][cadom_i][0]; // + (1 - settings->RyRParams_NP[0]) / (1 + pow(0.008 / Ca_cis_observed, 6));
		Inact_ss = (midlev / (1 + pow(Ca_sr_observed / 4,4)) + (settings->RyRParams_NP[seg_i][cadom_i][9] - midlev) / (1 + pow(Ca_sr_observed / settings->RyRParams_NP[seg_i][cadom_i][1], settings->RyRParams_NP[seg_i][cadom_i][2])));
		tau_Inact = settings->RyRParams_NP[seg_i][cadom_i][10] + settings->RyRParams_NP[seg_i][cadom_i][12] / (1 + pow(settings->RyRParams_NP[seg_i][cadom_i][1] / Ca_sr_observed, settings->RyRParams_NP[seg_i][cadom_i][2]));
				
		double tau_Flec = settings->RyRParams_NP[seg_i][cadom_i][24];
		double Flec_Block_ss = 0;
		if(settings->Flec > 0)
		{
			Flec_Block_ss = 1.0 / (1 + pow(settings->RyRParams_NP[seg_i][cadom_i][22] / settings->Flec, settings->RyRParams_NP[seg_i][cadom_i][23]));
		}
		
		//Methodology to clamp RyR2
		if(settings->clamp_RyR_Po >= -0.1 && settings->clamp_RyR_Po <= 1) 
		{
			if(settings->numRyR_Po_Indices < 0 || GenPurpose::occursInList(settings->RyR_Po_Indices, settings->numRyR_Po_Indices, (seg_i - 1) * settings->N_CaDomains + cadom_i))
			{
				if(settings->clamp_RyR_Po >= 0)
				{
					Act_ss = settings->clamp_RyR_Po;
				
					if(settings->clamp_RyR_Po >= 0.3) 
					{
						Inact_ss = 0;
						Flec_Block_ss = 0;
						segments[seg_i].caUnits[cadom_i].Jrel = segments[seg_i].caUnits[cadom_i].Jrel * 2;
					}	
					exp_Ca_cis = 1.0 / (1 + exp(-(Ca_cis_observed - settings->RyRParams_NP[seg_i][cadom_i][4])/settings->RyRParams_NP[seg_i][cadom_i][5]));
					tau_Act = settings->RyRParams_NP[seg_i][cadom_i][7] + settings->RyRParams_NP[seg_i][cadom_i][8] * kCaSR * (1 - exp_Ca_cis);
				}
				else
				{
					Act_ss = (-10 * settings->clamp_RyR_Po) * Act_ss;
				}
			}
		}

		k_12 = Act_ss / tau_Act;
		k_21 = (1 - Act_ss) / tau_Act;
		double k_43 = Act_ss_low / tau_Act_low;
		double k_34 = (1 - Act_ss_low) / tau_Act_low;
		double k_23 = Inact_ss / tau_Inact;
		double k_32 = (1 - Inact_ss) / tau_Inact;
		k_41 = 0.1 * k_32;
		k_14 = (k_12 * k_23 * k_34 * k_41) / (k_43 * k_32 * k_21);
		
		double k_26 = Flec_Block_ss / tau_Flec;
		double k_62 = (1 - Flec_Block_ss) / tau_Flec;
		
		double dR = (k_41*segments[seg_i].caUnits[cadom_i].RyRState_NP[3]-k_14*segments[seg_i].caUnits[cadom_i].RyRState_NP[0])-(k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[0]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[1]);   // R
		double dO = (k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[0]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[1])-(k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[1]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[2]) + k_62 * segments[seg_i].caUnits[cadom_i].RyRState_NP[5] - k_26 * segments[seg_i].caUnits[cadom_i].RyRState_NP[1]; // O
		double dI = (k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[1]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[2])-(k_34*segments[seg_i].caUnits[cadom_i].RyRState_NP[2]-k_43*segments[seg_i].caUnits[cadom_i].RyRState_NP[3]);   // I
		double dCI = k_14*segments[seg_i].caUnits[cadom_i].RyRState_NP[0] - k_41*segments[seg_i].caUnits[cadom_i].RyRState_NP[3] + k_34*segments[seg_i].caUnits[cadom_i].RyRState_NP[2] - k_43*segments[seg_i].caUnits[cadom_i].RyRState_NP[3];
		
		double dR_B = (k_41*segments[seg_i].caUnits[cadom_i].RyRState_NP[7]-k_14*segments[seg_i].caUnits[cadom_i].RyRState_NP[4])-(k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[4]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[5]);   // R
		double dO_B = (k_12*segments[seg_i].caUnits[cadom_i].RyRState_NP[4]-k_21*segments[seg_i].caUnits[cadom_i].RyRState_NP[5])-(k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[5]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[6]) - k_62 * segments[seg_i].caUnits[cadom_i].RyRState_NP[5] + k_26 * segments[seg_i].caUnits[cadom_i].RyRState_NP[1]; // O
		double dI_B = (k_23*segments[seg_i].caUnits[cadom_i].RyRState_NP[5]-k_32*segments[seg_i].caUnits[cadom_i].RyRState_NP[6])-(k_34*segments[seg_i].caUnits[cadom_i].RyRState_NP[6]-k_43*segments[seg_i].caUnits[cadom_i].RyRState_NP[7]);   // I
		//double dCI = k_14*segments[seg_i].caUnits[cadom_i].RyRState_NP[0] - k_41*segments[seg_i].caUnits[cadom_i].RyRState_NP[3] + k_34*segments[seg_i].caUnits[cadom_i].RyRState_NP[2] - k_43*segments[seg_i].caUnits[cadom_i].RyRState_NP[3];
			
		segments[seg_i].caUnits[cadom_i].RyRState_NP[0] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[0], dR, dt);	
		segments[seg_i].caUnits[cadom_i].RyRState_NP[1] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[1], dO, dt);
		segments[seg_i].caUnits[cadom_i].RyRState_NP[2] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[2], dI, dt);
		segments[seg_i].caUnits[cadom_i].RyRState_NP[3] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[3], dCI, dt);
		segments[seg_i].caUnits[cadom_i].RyRState_NP[4] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[4], dR_B, dt);
		segments[seg_i].caUnits[cadom_i].RyRState_NP[5] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[5], dO_B, dt);
		segments[seg_i].caUnits[cadom_i].RyRState_NP[6] = check_State_Change(segments[seg_i].caUnits[cadom_i].RyRState_NP[6], dI_B, dt);
		
		if(segments[seg_i].caUnits[cadom_i].RyRState_NP[0] < 0) { segments[seg_i].caUnits[cadom_i].RyRState_NP[3] += segments[seg_i].caUnits[cadom_i].RyRState_NP[0]; segments[seg_i].caUnits[cadom_i].RyRState_NP[0] = 0; } // Compensated by State 3
		if(segments[seg_i].caUnits[cadom_i].RyRState_NP[1] < 0) { segments[seg_i].caUnits[cadom_i].RyRState_NP[2] += segments[seg_i].caUnits[cadom_i].RyRState_NP[1]; segments[seg_i].caUnits[cadom_i].RyRState_NP[1] = 0; }
		if(segments[seg_i].caUnits[cadom_i].RyRState_NP[5] < 0) { segments[seg_i].caUnits[cadom_i].RyRState_NP[6] += segments[seg_i].caUnits[cadom_i].RyRState_NP[5]; segments[seg_i].caUnits[cadom_i].RyRState_NP[5] = 0; }
		
		segments[seg_i].caUnits[cadom_i].RyRState_NP[7] = 1.0-segments[seg_i].caUnits[cadom_i].RyRState_NP[0]-segments[seg_i].caUnits[cadom_i].RyRState_NP[1]-segments[seg_i].caUnits[cadom_i].RyRState_NP[2]-segments[seg_i].caUnits[cadom_i].RyRState_NP[3]-segments[seg_i].caUnits[cadom_i].RyRState_NP[4]-segments[seg_i].caUnits[cadom_i].RyRState_NP[5]-segments[seg_i].caUnits[cadom_i].RyRState_NP[6];
		if(segments[seg_i].caUnits[cadom_i].RyRState_NP[7] < 0)
		{		
			segments[seg_i].caUnits[cadom_i].RyRState_NP[3] += segments[seg_i].caUnits[cadom_i].RyRState_NP[7];
			segments[seg_i].caUnits[cadom_i].RyRState_NP[7] = 0;
		}
		
		double RyR_SEP_SS = 1.0 / (1.0 + exp((Ca_cis_observed - 1.8E-3) / 1E-6)); // 1 / (1 + pow(Ca_cis_observed / 2.75E-4, 20));
		double tau_RyR_SEP = 60;
		segments[seg_i].caUnits[cadom_i].RyR_SEP = check_State_Change(segments[seg_i].caUnits[cadom_i].RyR_SEP, (RyR_SEP_SS - segments[seg_i].caUnits[cadom_i].RyR_SEP) / tau_RyR_SEP, dt);
		
		double dRyR_Int = 1 * segments[seg_i].caUnits[cadom_i].Jrel - 0.1 / (1 + exp((segments[seg_i].caUnits[cadom_i].Jrel - 1.5E-3) / 1E-4)) * segments[seg_i].caUnits[cadom_i].RyR_Integr;
		segments[seg_i].caUnits[cadom_i].RyR_Integr = check_State_Change(segments[seg_i].caUnits[cadom_i].RyR_Integr, dRyR_Int, dt);
	}

	if(settings->SERCA_Distribution_Scaling[seg_i][cadom_i] > 1E-12)
	{	
		// SERCA
		double tval_Kmr = pow(segments[seg_i].caUnits[cadom_i].Ca_sr / data->Kmr, data->hillSRCaP);
		double tval_Kmf_NP = pow(segments[seg_i].caUnits[cadom_i].Ca_i / data->Kmf_NP, data->hillSRCaP);
		double Jup_NP = (1 - settings->IupB) * (settings->SERCA_Distribution_Scaling[seg_i][cadom_i] * settings->N_Segments * settings->N_CaDomains) * data->Vmax_SRCaP * (tval_Kmf_NP - tval_Kmr) / (1 + tval_Kmf_NP + tval_Kmr);	//double Jup_P = (1 - settings->IupB) * data->Vmax_SRCaP * (tval_Kmf_P - tval_Kmr) / (1 + tval_Kmf_P + tval_Kmr);
		segments[seg_i].caUnits[cadom_i].Jup = Jup_NP;
	}
}

double HAMM::Buff_Ca_DoubleBuff(double ca_t, double buff1_tot, double buff2_tot, double Kmbuff1, double Kmbuff2)
{
	double val;
	double b_buff = buff1_tot+buff2_tot-ca_t+Kmbuff1+Kmbuff2;
	double c_buff = Kmbuff1*Kmbuff2 -ca_t*(Kmbuff1+Kmbuff2)+buff2_tot*Kmbuff1+buff1_tot*Kmbuff2;
	double d_buff = -Kmbuff1*Kmbuff2*ca_t;      
    
	double sq_bbc_val = 2 * sqrt(b_buff*b_buff-3*c_buff);
	val = (sq_bbc_val/3.0)*cos(acos((9*b_buff*c_buff-2*b_buff*b_buff*b_buff-27*d_buff)/((b_buff*b_buff-3*c_buff)*sq_bbc_val))/3)-(b_buff/3);

	return val;
}

HAMM* HAMM::clone()
{
	HAMM* cl = new HAMM();
		cl->init(data, settings);
		cl->Vm = Vm;
		
		for(int ii=0; ii<settings->N_Segments; ii++)
		{	
			// Loop through the two membrane sides
			for(int jj=0; jj<2; jj++)
			{
				cl->segments[ii].membrane[jj].m_Na = segments[ii].membrane[jj].m_Na; cl->segments[ii].membrane[jj].h_Na = segments[ii].membrane[jj].h_Na; cl->segments[ii].membrane[jj].j_Na = segments[ii].membrane[jj].j_Na; 
				cl->segments[ii].membrane[jj].mL_Na = segments[ii].membrane[jj].mL_Na; cl->segments[ii].membrane[jj].hL_Na = segments[ii].membrane[jj].hL_Na; cl->segments[ii].membrane[jj].xr = segments[ii].membrane[jj].xr; 
				cl->segments[ii].membrane[jj].to_a = segments[ii].membrane[jj].to_a; cl->segments[ii].membrane[jj].to_if = segments[ii].membrane[jj].to_if; cl->segments[ii].membrane[jj].kur_a = segments[ii].membrane[jj].kur_a; 
				cl->segments[ii].membrane[jj].kur_i = segments[ii].membrane[jj].kur_i; cl->segments[ii].membrane[jj].xs_NP = segments[ii].membrane[jj].xs_NP; cl->segments[ii].membrane[jj].xs_P = segments[ii].membrane[jj].xs_P;
			
				cl->segments[ii].membrane[jj].Buff_Na_sl = segments[ii].membrane[jj].Buff_Na_sl;
				cl->segments[ii].membrane[jj].Buff_Ca_SLLj = segments[ii].membrane[jj].Buff_Ca_SLLj; cl->segments[ii].membrane[jj].Buff_Ca_SLLsl = segments[ii].membrane[jj].Buff_Ca_SLLsl;
				cl->segments[ii].membrane[jj].Buff_Ca_SLHj = segments[ii].membrane[jj].Buff_Ca_SLHj; cl->segments[ii].membrane[jj].Buff_Ca_SLHsl = segments[ii].membrane[jj].Buff_Ca_SLHsl;
				cl->segments[ii].membrane[jj].Buff_Ca_EGTA_sl = segments[ii].membrane[jj].Buff_Ca_EGTA_sl;
								
				cl->segments[ii].membrane[jj].Ca_junc = segments[ii].membrane[jj].Ca_junc;
				cl->segments[ii].membrane[jj].Ca_sl = segments[ii].membrane[jj].Ca_sl;
				cl->segments[ii].membrane[jj].Na_sl = segments[ii].membrane[jj].Na_sl;
			}
			
			// Loop through the cytosolic Ca2+ domains
			for(int jj=0; jj<settings->N_CaDomains; jj++)
			{
				cl->segments[ii].caUnits[jj].Buff_Na_junc = segments[ii].caUnits[jj].Buff_Na_junc;
				cl->segments[ii].caUnits[jj].Na_srs = segments[ii].caUnits[jj].Na_srs;
				cl->segments[ii].caUnits[jj].Na_i = segments[ii].caUnits[jj].Na_i;
				
				cl->segments[ii].caUnits[jj].RyRState_NP[0] = segments[ii].caUnits[jj].RyRState_NP[0]; cl->segments[ii].caUnits[jj].RyRState_NP[1] = segments[ii].caUnits[jj].RyRState_NP[1]; cl->segments[ii].caUnits[jj].RyRState_NP[2] = segments[ii].caUnits[jj].RyRState_NP[2]; 
				cl->segments[ii].caUnits[jj].RyRState_NP[3] = segments[ii].caUnits[jj].RyRState_NP[3]; cl->segments[ii].caUnits[jj].RyRState_NP[4] = segments[ii].caUnits[jj].RyRState_NP[4]; cl->segments[ii].caUnits[jj].RyRState_NP[5] = segments[ii].caUnits[jj].RyRState_NP[5]; 
				cl->segments[ii].caUnits[jj].RyRState_NP[6] = segments[ii].caUnits[jj].RyRState_NP[6]; cl->segments[ii].caUnits[jj].RyRState_NP[7] = segments[ii].caUnits[jj].RyRState_NP[7];
				
				cl->segments[ii].caUnits[jj].RyR_SEP = segments[ii].caUnits[jj].RyR_SEP;
				cl->segments[ii].caUnits[jj].RyR_Integr = segments[ii].caUnits[jj].RyR_Integr;

				for(int abcd = 0; abcd<8; abcd++)
				{
					cl->segments[ii].caUnits[jj].ICaLState_NP[abcd] = segments[ii].caUnits[jj].ICaLState_NP[abcd];
				}				
				
				cl->segments[ii].caUnits[jj].Buff_Ca_TnCL = segments[ii].caUnits[jj].Buff_Ca_TnCL;
				cl->segments[ii].caUnits[jj].Buff_Ca_TnCHc = segments[ii].caUnits[jj].Buff_Ca_TnCHc;
				cl->segments[ii].caUnits[jj].Buff_Ca_TnCHm = segments[ii].caUnits[jj].Buff_Ca_TnCHm;
				cl->segments[ii].caUnits[jj].Buff_Ca_CaM = segments[ii].caUnits[jj].Buff_Ca_CaM;
				cl->segments[ii].caUnits[jj].Buff_Ca_Myosin_ca = segments[ii].caUnits[jj].Buff_Ca_Myosin_ca;
				cl->segments[ii].caUnits[jj].Buff_Ca_Myosin_mg = segments[ii].caUnits[jj].Buff_Ca_Myosin_mg;
				cl->segments[ii].caUnits[jj].Buff_Ca_SRB = segments[ii].caUnits[jj].Buff_Ca_SRB;
				cl->segments[ii].caUnits[jj].Buff_Ca_CSQN = segments[ii].caUnits[jj].Buff_Ca_CSQN;
				cl->segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt = segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt;
		
				cl->segments[ii].caUnits[jj].Ca_i = segments[ii].caUnits[jj].Ca_i;
				cl->segments[ii].caUnits[jj].Ca_sr = segments[ii].caUnits[jj].Ca_sr;
				
				cout << "Clone RyR_State = [" << cl->segments[ii].caUnits[jj].RyRState_NP[0] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[1] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[2] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[3] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[4] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[5] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[6] << "," << cl->segments[ii].caUnits[jj].RyRState_NP[7] << "]" << endl;
			}
			
			// Other stuff
			cl->segments[ii].K_i = segments[ii].K_i;
		}
		
	return cl;
}

void HAMM::outputData(FILE* outputfile, int levelOfElectroDetail, int levelOfSignalingDetail, bool binaryOutput)
{     
	// Minimum Level of Detail:
	//   Vm 
	if(levelOfElectroDetail == 1)
	{
		fprintf(outputfile, "%-08e\t", Vm);		
	}

	// Average Level of Detail (1):
	if(levelOfElectroDetail == 2)
	{		
		double Irel_adjust = Jrel + Jleak;
		fprintf(outputfile,"%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t",Vm,Ca_i,Ca_sr,(ICaL_junc+ICaL_sl),Irel_adjust,IKAch,INaCa_junc+INaCa_sl,Jup,Ca_sl,ITo1,Na_i,Ca_junc,INaK_junc+INaK_sl, IKp, IKr, IK1, K_i);
	}

	if(levelOfElectroDetail == 21)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;	
		double dVdt = caiont + naiont + kiont_alt + clont;
		double Avg_Ca_sr_tot = 0;
		double totnumdoms = (1.0 * settings->N_Segments * settings->N_CaDomains);
		for(int i=0; i<settings->N_Segments; i++) 
		{
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				Avg_Ca_sr_tot += (1.0 / totnumdoms) * (segments[i].caUnits[j].Ca_sr + segments[i].caUnits[j].Buff_Ca_CSQN);
			}
		}
		double Irel_adjust = Jrel + Jleak;
		
		fprintf(outputfile,"%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t%-08e\t",Vm,1000*Ca_i,Ca_sr,dVdt,Irel_adjust,ICab_junc+ICab_sl,INaCa_junc+INaCa_sl,Jup,Ca_sl,IpCa_junc+IpCa_sl,Na_i,Ca_junc,INaK_junc+INaK_sl, Avg_Ca_sr_tot, ICaL_junc+ICaL_sl, INaCa_junc, K_i);
	}
		
	if(levelOfElectroDetail == 4)
	{
		fprintf(outputfile, "%-06e\t%-06e\t", Vm, 1000*Ca_i);
	}

	if(levelOfElectroDetail == 41)
	{
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i,(ICaL_junc+ICaL_sl));
	}
	
	if(levelOfElectroDetail == 42)
	{
		double Irel_adjust = Jleak + Jrel;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i,(ICaL_junc+ICaL_sl), Ca_sr, Irel_adjust);
	}
	
	if(levelOfElectroDetail == 44)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, caiont + naiont + kiont_alt + clont);
	}
	
	if(levelOfElectroDetail == 43)
	{
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i,Istim);
	}

	if(levelOfElectroDetail == 45)
	{
		double Irel_adjust = Jleak + Jrel;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, Ca_i, Ca_sr, Ca_sl, Ca_junc, Irel_adjust, (ICaL_junc+ICaL_sl));
		for(int i=0; i<settings->N_Segments; i++) 
		{
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				fprintf(outputfile, "%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel);
			}
		}
	}
	
	if(levelOfElectroDetail == 46)
	{
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, INa_junc + INa_sl, INaK_junc + INaK_sl, INaCa_junc+INaCa_sl, ICaL_junc+ICaL_sl, ITo1, IK1, IKp, IKr, IKs, IKur, IClb, IClCa_junc+IClCa_sl, IpCa_junc + IpCa_sl);
	}
	
	if(levelOfElectroDetail == 47)
	{
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, INa_junc + INa_sl, segments[0].membrane[0].m_Na, segments[0].membrane[0].h_Na, segments[0].membrane[0].j_Na);
	}
	
	if(levelOfElectroDetail == 48)
	{
		double Irel_adjust = Jleak + Jrel;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, Ca_i, Ca_sr, Ca_sl, Ca_junc, Irel_adjust, Jup, INaCa_junc+INaCa_sl, ICaL_junc+ICaL_sl);
		for(int i=0; i<settings->N_Segments; i++) 
		{
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel, segments[i].caUnits[j].RyRState_NP[0], segments[i].caUnits[j].RyRState_NP[1], segments[i].caUnits[j].RyRState_NP[2], segments[i].caUnits[j].Jup, segments[i].caUnits[j].Ca_srs);
			}
		}
	}
	
	if(levelOfElectroDetail == 49)
	{
		double Irel_adjust = Jleak + Jrel;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, Ca_i, Ca_sr, Ca_sl, Ca_junc, Irel_adjust, Jup);		
	}
	
	if(levelOfElectroDetail == 5)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;
		double Irel_adjust = Jleak + Jrel;
		double ICaL_tot = ICaL_junc + ICaL_sl;
		
		if(!binaryOutput)
		{
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, Ca_sr, Ca_sl, Buff_Ca_EGTA_cyt, Irel_adjust, ICaL_tot, caiont + naiont + kiont_alt + clont);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{
					fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel);
				}
			}
		}
		else
		{
			size_t doubles = sizeof(double);
			double Cai_x_1000 = 1000 * Ca_i;
			double dVdt = caiont + naiont + kiont_alt + clont;
			fwrite(&(Vm), doubles, 1, outputfile); fwrite(&(Cai_x_1000), doubles, 1, outputfile); fwrite(&(Ca_sr), doubles, 1, outputfile); fwrite(&(Ca_sl), doubles, 1, outputfile); 
			fwrite(&(Buff_Ca_EGTA_cyt), doubles, 1, outputfile); fwrite(&(Irel_adjust), doubles, 1, outputfile); fwrite(&(ICaL_tot), doubles, 1, outputfile); fwrite(&(dVdt), doubles, 1, outputfile); 
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{									
					fwrite(&(segments[i].caUnits[j].Ca_i), doubles, 1, outputfile); //fwrite(&(segments[i].caUnits[j].Ca_sr), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Jrel), doubles, 1, outputfile); //fwrite(&(segments[i].caUnits[j].Ca_srs), doubles, 1, outputfile); 
				}
			}
		}
	}
	
	if(levelOfElectroDetail == 51)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;
		
		double TotCellCa = 0;
		double totnumdoms = (1.0 * settings->N_Segments * settings->N_CaDomains);
		for(int i=0; i<settings->N_Segments; i++) 
		{		
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				TotCellCa += (data->vmyo / totnumdoms) / data->vcell * (segments[i].caUnits[j].Ca_i + segments[i].caUnits[j].Buff_Ca_TnCL + segments[i].caUnits[j].Buff_Ca_TnCHc + segments[i].caUnits[j].Buff_Ca_TnCHm + segments[i].caUnits[j].Buff_Ca_CaM + segments[i].caUnits[j].Buff_Ca_Myosin_ca + segments[i].caUnits[j].Buff_Ca_Myosin_mg + segments[i].caUnits[j].Buff_Ca_SRB);
			}
		}
		double Irel_adjust = Jleak + Jrel;
		
		if(!binaryOutput)
		{	
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, Ca_sr, Ca_sl, Buff_Ca_EGTA_cyt, Irel_adjust, (ICaL_junc+ICaL_sl), caiont + naiont + kiont_alt + clont);
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[0].caUnits[0].ICaLState_NP[0], segments[0].caUnits[0].ICaLState_NP[1], segments[0].caUnits[0].ICaLState_NP[2], segments[0].caUnits[0].ICaLState_NP[3], segments[0].caUnits[0].ICaLState_NP[4], segments[0].caUnits[0].ICaLState_NP[5], segments[0].caUnits[0].ICaLState_NP[6], segments[0].membrane[0].Ca_sl,segments[0].caUnits[0].Ca_srs);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{
					fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel, segments[i].caUnits[j].Ca_srs);
				}
			}
		}
		else
		{
			size_t doubles = sizeof(double);
			double Cai_x_1000 = 1000 * Ca_i;
			double ICaL_tot = (ICaL_junc+ICaL_sl);
			double dVdt = caiont + naiont + kiont_alt + clont;
			fwrite(&(Vm), doubles, 1, outputfile); fwrite(&(Cai_x_1000), doubles, 1, outputfile); fwrite(&(Ca_sr), doubles, 1, outputfile); fwrite(&(Ca_sl), doubles, 1, outputfile); 
			fwrite(&(Buff_Ca_EGTA_cyt), doubles, 1, outputfile); fwrite(&(Irel_adjust), doubles, 1, outputfile); fwrite(&(ICaL_tot), doubles, 1, outputfile); fwrite(&(dVdt), doubles, 1, outputfile); 
			fwrite(segments[0].caUnits[0].RyRState_NP, doubles, 4, outputfile);
			fwrite(segments[0].caUnits[1].RyRState_NP, doubles, 4, outputfile); fwrite(&(segments[0].caUnits[0].Ca_srs), doubles, 1, outputfile);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{									
					fwrite(&(segments[i].caUnits[j].Ca_i), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Ca_sr), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Jrel), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Ca_srs), doubles, 1, outputfile); 
				}
			}
		}
	}
	
	if(levelOfElectroDetail == 52)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;
		
		double TotCellCa = 0;
		double totnumdoms = (1.0 * settings->N_Segments * settings->N_CaDomains);
		for(int i=0; i<settings->N_Segments; i++) 
		{
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				TotCellCa += (data->vmyo / totnumdoms) / data->vcell * (segments[i].caUnits[j].Ca_i + segments[i].caUnits[j].Buff_Ca_TnCL + segments[i].caUnits[j].Buff_Ca_TnCHc + segments[i].caUnits[j].Buff_Ca_TnCHm + segments[i].caUnits[j].Buff_Ca_CaM + segments[i].caUnits[j].Buff_Ca_Myosin_ca + segments[i].caUnits[j].Buff_Ca_Myosin_mg + segments[i].caUnits[j].Buff_Ca_SRB);
			}
		}
		double Irel_adjust = Jleak + Jrel;
		
		if(!binaryOutput)
		{	
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, Ca_sr, Ca_sl, Buff_Ca_EGTA_cyt, Irel_adjust, (ICaL_junc+ICaL_sl), caiont + naiont + kiont_alt + clont);
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[0].caUnits[0].ICaLState_NP[0], segments[0].caUnits[0].ICaLState_NP[1], segments[0].caUnits[0].ICaLState_NP[2], segments[0].caUnits[0].ICaLState_NP[3], segments[0].caUnits[0].ICaLState_NP[4], segments[0].caUnits[0].ICaLState_NP[5], segments[0].caUnits[0].ICaLState_NP[6], segments[0].membrane[0].Ca_sl,segments[0].caUnits[0].Ca_srs);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{
					fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel, segments[i].caUnits[j].Ca_srs, segments[i].caUnits[j].RyRState_NP[0], segments[i].caUnits[j].RyRState_NP[1], segments[i].caUnits[j].RyRState_NP[2], segments[i].caUnits[j].RyRState_NP[3]);
				}
			}
		}
		else
		{
			size_t doubles = sizeof(double);
			double Cai_x_1000 = 1000 * Ca_i;
			double ICaL_tot = (ICaL_junc+ICaL_sl);
			double dVdt = caiont + naiont + kiont_alt + clont;
			fwrite(&(Vm), doubles, 1, outputfile); fwrite(&(Cai_x_1000), doubles, 1, outputfile); fwrite(&(Ca_sr), doubles, 1, outputfile); fwrite(&(Ca_sl), doubles, 1, outputfile); 
			fwrite(&(Buff_Ca_EGTA_cyt), doubles, 1, outputfile); fwrite(&(Irel_adjust), doubles, 1, outputfile); fwrite(&(ICaL_tot), doubles, 1, outputfile); fwrite(&(dVdt), doubles, 1, outputfile); 
			fwrite(segments[0].caUnits[0].RyRState_NP, doubles, 4, outputfile);
			fwrite(segments[0].caUnits[1].RyRState_NP, doubles, 4, outputfile); fwrite(&(segments[0].caUnits[0].Ca_srs), doubles, 1, outputfile);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{									
					fwrite(&(segments[i].caUnits[j].Ca_i), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Ca_sr), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Jrel), doubles, 1, outputfile); fwrite(&(segments[i].caUnits[j].Ca_srs), doubles, 1, outputfile); 
					fwrite(&(segments[i].caUnits[j].RyRState_NP), doubles, 4, outputfile);
				}
			}
		}
	}
	
	if(levelOfElectroDetail == 53)
	{
		double kiont_alt = IKr+IKs+IK1+IKur+IKp-2*(INaK_junc+INaK_sl)+ITo1;
		if(!settings->applyVoltageClamp) kiont_alt +=Istim;
		
		double TotCellCa = 0;
		double totnumdoms = (1.0 * settings->N_Segments * settings->N_CaDomains);
		for(int i=0; i<settings->N_Segments; i++) 
		{
			for(int j=0; j<settings->N_CaDomains; j++)
			{
				TotCellCa += (data->vmyo / totnumdoms) / data->vcell * (segments[i].caUnits[j].Ca_i + segments[i].caUnits[j].Buff_Ca_TnCL + segments[i].caUnits[j].Buff_Ca_TnCHc + segments[i].caUnits[j].Buff_Ca_TnCHm + segments[i].caUnits[j].Buff_Ca_CaM + segments[i].caUnits[j].Buff_Ca_Myosin_ca + segments[i].caUnits[j].Buff_Ca_Myosin_mg + segments[i].caUnits[j].Buff_Ca_SRB);
			}
		}
		double Irel_adjust = Jleak + Jrel;
		
		if(!binaryOutput)
		{	
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, Ca_sr, Ca_sl, Buff_Ca_EGTA_cyt, Irel_adjust, (ICaL_junc+ICaL_sl), caiont + naiont + kiont_alt + clont);
			fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[0].caUnits[0].ICaLState_NP[0], segments[0].caUnits[0].ICaLState_NP[1], segments[0].caUnits[0].ICaLState_NP[2], segments[0].caUnits[0].ICaLState_NP[3], segments[0].caUnits[0].ICaLState_NP[4], segments[0].caUnits[0].ICaLState_NP[5], segments[0].caUnits[0].ICaLState_NP[6], segments[0].membrane[0].Ca_sl,segments[0].caUnits[0].Ca_srs);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{
					fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", segments[i].caUnits[j].Ca_i, segments[i].caUnits[j].Ca_sr, segments[i].caUnits[j].Jrel, segments[i].caUnits[j].Ca_srs, segments[i].caUnits[j].RyR_Integr);
				}
			}
		}
		else
		{
			size_t doubles = sizeof(double);
			double Cai_x_1000 = 1000 * Ca_i;
			double ICaL_tot = (ICaL_junc+ICaL_sl);
			double dVdt = caiont + naiont + kiont_alt + clont;
			fwrite(&(Vm), doubles, 1, outputfile); fwrite(&(Cai_x_1000), doubles, 1, outputfile); fwrite(&(Ca_sr), doubles, 1, outputfile); fwrite(&(Ca_sl), doubles, 1, outputfile); 
			fwrite(&(Buff_Ca_EGTA_cyt), doubles, 1, outputfile); fwrite(&(Irel_adjust), doubles, 1, outputfile); fwrite(&(ICaL_tot), doubles, 1, outputfile); fwrite(&(dVdt), doubles, 1, outputfile); 
			fwrite(segments[0].caUnits[0].RyRState_NP, doubles, 4, outputfile);
			fwrite(segments[0].caUnits[1].RyRState_NP, doubles, 4, outputfile); fwrite(&(segments[0].caUnits[0].Ca_srs), doubles, 1, outputfile);
			for(int i=0; i<settings->N_Segments; i+=settings->segOutputRes) 
			{
				for(int j=0; j<settings->N_CaDomains; j+=settings->domOutputRes)
				{									
					fwrite(&(segments[i].caUnits[j].Ca_i), doubles, 1, outputfile); 
					fwrite(&(segments[i].caUnits[j].Ca_sr), doubles, 1, outputfile); 
					fwrite(&(segments[i].caUnits[j].Jrel), doubles, 1, outputfile); 
					fwrite(&(segments[i].caUnits[j].Ca_srs), doubles, 1, outputfile); 
					fwrite(&(segments[i].caUnits[j].RyR_Integr), doubles, 1, outputfile); 
				}
			}
		}
	}
	
	if(levelOfElectroDetail == 9)
	{		
		double Irel_adjust = Jleak + Jrel;
		fprintf(outputfile, "%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t%-04e\t", Vm, 1000*Ca_i, Ca_sr, Ca_sl, Ca_junc, Irel_adjust, INa_junc+INa_sl, (ICaL_junc+ICaL_sl), IKr, IKs, ITo1, IKur, IK1, INaK_junc+INaK_sl, INaCa_junc+INaCa_sl, Na_i);
	}

	// Full detail
	//    Also used to store X0, so don't alter this!
	if(levelOfElectroDetail == 10)
	{
		fprintf(outputfile, "%-08e\t", Vm);
		
		for(int ii=0; ii<settings->N_Segments; ii++)
		{	
			// Loop through the two membrane sides
			for(int jj=0; jj<2; jj++)
			{
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].membrane[jj].m_Na, segments[ii].membrane[jj].h_Na, segments[ii].membrane[jj].j_Na); 
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].membrane[jj].mL_Na, segments[ii].membrane[jj].hL_Na, segments[ii].membrane[jj].xr); 
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].membrane[jj].to_a, segments[ii].membrane[jj].to_if, segments[ii].membrane[jj].kur_a); 
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].membrane[jj].kur_i, segments[ii].membrane[jj].xs_NP, segments[ii].membrane[jj].xs_P);
			
				fprintf(outputfile, "%-08e\t", segments[ii].membrane[jj].Buff_Na_sl);
				fprintf(outputfile, "%-08e\t%-08e\t", segments[ii].membrane[jj].Buff_Ca_SLLj, segments[ii].membrane[jj].Buff_Ca_SLLsl);
				fprintf(outputfile, "%-08e\t%-08e\t", segments[ii].membrane[jj].Buff_Ca_SLHj, segments[ii].membrane[jj].Buff_Ca_SLHsl);
				fprintf(outputfile, "%-08e\t", segments[ii].membrane[jj].Buff_Ca_EGTA_sl);
				
				fprintf(outputfile, "%-08e\t%-08e\t", segments[ii].membrane[jj].Ca_junc, segments[ii].membrane[jj].Ca_sl);
				fprintf(outputfile, "%-08e\t", segments[ii].membrane[jj].Na_sl);				
			}
			
			// Loop through the cytosolic Ca2+ domains
			for(int jj=0; jj<settings->N_CaDomains; jj++)
			{
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t%-08e\t",segments[ii].caUnits[jj].RyRState_NP[0], segments[ii].caUnits[jj].RyRState_NP[1], segments[ii].caUnits[jj].RyRState_NP[2], segments[ii].caUnits[jj].RyRState_NP[3]);
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t%-08e\t",segments[ii].caUnits[jj].RyRState_NP[4], segments[ii].caUnits[jj].RyRState_NP[5], segments[ii].caUnits[jj].RyRState_NP[6], segments[ii].caUnits[jj].RyRState_NP[7]);
				
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].caUnits[jj].Buff_Ca_TnCL, segments[ii].caUnits[jj].Buff_Ca_TnCHc, segments[ii].caUnits[jj].Buff_Ca_TnCHm);
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].caUnits[jj].Buff_Ca_CaM, segments[ii].caUnits[jj].Buff_Ca_Myosin_ca, segments[ii].caUnits[jj].Buff_Ca_Myosin_mg);
				fprintf(outputfile, "%-08e\t%-08e\t", segments[ii].caUnits[jj].Buff_Ca_SRB, segments[ii].caUnits[jj].Buff_Ca_CSQN);
		
				fprintf(outputfile, "%-08e\t%-08e\t", segments[ii].caUnits[jj].Ca_i, segments[ii].caUnits[jj].Ca_sr);
				fprintf(outputfile, "%-08e\t", segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt);
				
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].caUnits[jj].Ca_srs, segments[ii].caUnits[jj].Buff_Ca_SLLsrs, segments[ii].caUnits[jj].Buff_Ca_SLHsrs);
				fprintf(outputfile, "%-08e\t", segments[ii].caUnits[jj].RyR_Integr);
				for(int abcd=0; abcd<8; abcd++)
				{
					fprintf(outputfile, "%-08e\t", segments[ii].caUnits[jj].ICaLState_NP[abcd]);
				}
				
				fprintf(outputfile, "%-08e\t%-08e\t%-08e\t", segments[ii].caUnits[jj].Na_i,segments[ii].caUnits[jj].Na_srs, segments[ii].caUnits[jj].Buff_Na_junc);
			}
			
			// Other stuff
			fprintf(outputfile, "%-08e\t", segments[ii].K_i);
		}
	}
}

void HAMM::processX0DataForMarkovModel(double* pX0Data, int sindex, double* MM, int num_states, int model_type, int num_channels)
{
	double statesum = 0;	
	for(int sk=0; sk<num_states; sk++) { MM[sk] = pX0Data[sindex+sk]; statesum += MM[sk]; }
	if(model_type == 3 && statesum < 1.1) 
	{
		cout << "Input is deterministic, target stochastic. Adapting to " << num_channels << " channels." << endl;
		statesum = 0; 
		for(int sk=0; sk<num_states-1; sk++) 
		{ 
			MM[sk] = max((int)floor(num_channels * MM[sk]),0); 
			statesum += MM[sk]; 
		}
		MM[num_states-1] = num_channels - statesum;
	}
	else
	{
		if(model_type < 3 && statesum > 2.1) // Deterministic model but stochastic input.
		{
			cout << "Input is stochastic, target deterministic. Scaling by " << statesum << "." << endl;
			for(int sk=0; sk<num_states; sk++)
			{
				MM[sk] = 1.0 * MM[sk] / statesum;
			}
		}
	}
}

void HAMM::processX0DataForMarkovModel(double* pX0Data, int sindex, int* MM, int num_states, int model_type, int num_channels)
{
	double statesum = 0;	
	for(int sk=0; sk<num_states; sk++) { MM[sk] = (int)(pX0Data[sindex+sk]); statesum += MM[sk]; }
	
	if(statesum != num_channels)
	{			
		double fact = (1.0 * num_channels) / statesum;
		int newsum = 0;
		
		cout << "WARNING: Num channels (" << num_channels << ") does not match state sum (" << statesum << "), scaling by " << fact << endl;
		GenPurpose::wait(5.0);
		
		for(int sk=0; sk<num_states-1; sk++)
		{
			MM[sk] = (int)(MM[sk] * fact);
			newsum += MM[sk];
		}
		MM[num_states-1] = num_channels - newsum;
	}
}

int HAMM::lengthMinimumStateVector()
{
	// Total number of variables:
	//		Global electro variables:	  1 + 1 x N_Segments (Vm + Ki)
	//		Membrane electro variables:  21 x 2 x N_Segments
	//		Local CaRU variables:        34 x N_Segments x N_CaDomains
	//		Global signaling variables:  00
	//									-------------
	//									 1 + 43 x N_Segments + 34 x N_Segments x N_CaDomains
	
	return 1 + 43 + 2 * 34;
}

void HAMM::loadX0Data(double* x0data, int x0length)
{
	int startat = -1;
	int numVarsPerMembr = 21;
	int numVarsPerCaDomain = 34;
	int numVarsPerSegment = 1 + 2 * numVarsPerMembr + numVarsPerCaDomain * settings->N_CaDomains;
	
	Vm = x0data[startat+1];
	
	if(x0length >= (1 + settings->N_Segments*numVarsPerSegment))
	{
		cout << "Using data from every individual Segment to populate " << settings->N_Segments << " Segments" << endl;
	}
	else
	{
		if(x0length >= 1 + numVarsPerSegment)
		{
			cout << "Using data from first Segment to populate all " << settings->N_Segments << " Segments" << endl;
			numVarsPerSegment = 0;
		}
		else
		{
			cout << "ERROR: X0 is of unsufficient length (" << x0length << "), minimum = " << numVarsPerSegment << endl;
			return;
		}
	}
	
	for(int ii=0; ii<settings->N_Segments; ii++)
	{	
		// Loop through the two membrane sides
		for(int jj=0; jj<2; jj++)
		{
			segments[ii].membrane[jj].m_Na = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+2];
			segments[ii].membrane[jj].h_Na = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+3];
			segments[ii].membrane[jj].j_Na = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+4];
			segments[ii].membrane[jj].mL_Na = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+5]; 
			segments[ii].membrane[jj].hL_Na = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+6]; 
			segments[ii].membrane[jj].xr = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+7]; 
			segments[ii].membrane[jj].to_a = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+8]; 
			segments[ii].membrane[jj].to_if = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+9]; 
			segments[ii].membrane[jj].kur_a = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+10]; 
			segments[ii].membrane[jj].kur_i = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+11];
			segments[ii].membrane[jj].xs_NP = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+12];
			segments[ii].membrane[jj].xs_P = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+13];

			segments[ii].membrane[jj].Buff_Na_sl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+14];
			segments[ii].membrane[jj].Buff_Ca_SLLj = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+15];
			segments[ii].membrane[jj].Buff_Ca_SLLsl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+16];
			segments[ii].membrane[jj].Buff_Ca_SLHj = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+17];
			segments[ii].membrane[jj].Buff_Ca_SLHsl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+18];
			segments[ii].membrane[jj].Buff_Ca_EGTA_sl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+19];
			
			segments[ii].membrane[jj].Ca_junc = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+20];
			segments[ii].membrane[jj].Ca_sl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+21];
			segments[ii].membrane[jj].Na_sl = x0data[ii*numVarsPerSegment+jj*numVarsPerMembr+startat+22];			
		}
		
		// Loop through the cytosolic Ca2+ domains
		for(int jj=0; jj<settings->N_CaDomains; jj++)
		{
			segments[ii].caUnits[jj].RyRState_NP[0] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+2]; 
			segments[ii].caUnits[jj].RyRState_NP[1] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+3];
			segments[ii].caUnits[jj].RyRState_NP[2] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+4];
			segments[ii].caUnits[jj].RyRState_NP[3] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+5];
			segments[ii].caUnits[jj].RyRState_NP[4] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+6]; 
			segments[ii].caUnits[jj].RyRState_NP[5] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+7];
			segments[ii].caUnits[jj].RyRState_NP[6] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+8];
			segments[ii].caUnits[jj].RyRState_NP[7] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+9];
			
			double sumRyR = segments[ii].caUnits[jj].RyRState_NP[0] + segments[ii].caUnits[jj].RyRState_NP[1] + segments[ii].caUnits[jj].RyRState_NP[2] + segments[ii].caUnits[jj].RyRState_NP[3] + segments[ii].caUnits[jj].RyRState_NP[4] + segments[ii].caUnits[jj].RyRState_NP[5] + segments[ii].caUnits[jj].RyRState_NP[6] + segments[ii].caUnits[jj].RyRState_NP[7];
			if(settings->RyRModelType == 3)
			{				
				if(fabs(sumRyR * (settings->N_Segments * settings->N_CaDomains) - settings->numRyRs) > 0.001)
				{
					// NOTE: The scaling from deterministic to stochastic is based on numRyR / (N_Segments * n_CaDomain), this does not take heterogeneity into account
					double fact = settings->numRyRs * 1.0 / (settings->N_Segments * settings->N_CaDomains);					
					cout << "Sum RyR in unit [" << ii << "," << jj << "] = " << sumRyR << ", rescaling to " << fact << ", old dist = [" << segments[ii].caUnits[jj].RyRState_NP[0] << "," << segments[ii].caUnits[jj].RyRState_NP[1] << "," << segments[ii].caUnits[jj].RyRState_NP[2] << "," << segments[ii].caUnits[jj].RyRState_NP[3] << "]";
					segments[ii].caUnits[jj].RyRState_NP[0] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[0]) / sumRyR);
					segments[ii].caUnits[jj].RyRState_NP[1] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[1]) / sumRyR);
					segments[ii].caUnits[jj].RyRState_NP[2] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[2]) / sumRyR);
					segments[ii].caUnits[jj].RyRState_NP[3] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[3]) / sumRyR);					
					segments[ii].caUnits[jj].RyRState_NP[4] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[4]) / sumRyR);					
					segments[ii].caUnits[jj].RyRState_NP[5] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[5]) / sumRyR);					
					segments[ii].caUnits[jj].RyRState_NP[6] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[6]) / sumRyR);					
					segments[ii].caUnits[jj].RyRState_NP[7] = (int)((fact * segments[ii].caUnits[jj].RyRState_NP[7]) / sumRyR);					
					cout << ", new dist = [" << segments[ii].caUnits[jj].RyRState_NP[0] << "," << segments[ii].caUnits[jj].RyRState_NP[1] << "," << segments[ii].caUnits[jj].RyRState_NP[2] << "," << segments[ii].caUnits[jj].RyRState_NP[3] << "]" << endl;
				}
			}
			else
			{
				if(settings->RyRModelType == 2 && sumRyR > 2.1)
				{
					cout << "Input is stochastic, target deterministic. Scaling by " << sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[0] = segments[ii].caUnits[jj].RyRState_NP[0] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[1] = segments[ii].caUnits[jj].RyRState_NP[1] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[2] = segments[ii].caUnits[jj].RyRState_NP[2] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[3] = segments[ii].caUnits[jj].RyRState_NP[3] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[4] = segments[ii].caUnits[jj].RyRState_NP[4] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[5] = segments[ii].caUnits[jj].RyRState_NP[5] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[6] = segments[ii].caUnits[jj].RyRState_NP[6] / sumRyR;
					segments[ii].caUnits[jj].RyRState_NP[7] = segments[ii].caUnits[jj].RyRState_NP[7] / sumRyR;
					cout << ", new dist = [" << segments[ii].caUnits[jj].RyRState_NP[0] << "," << segments[ii].caUnits[jj].RyRState_NP[1] << "," << segments[ii].caUnits[jj].RyRState_NP[2] << "," << segments[ii].caUnits[jj].RyRState_NP[3] << "]" << endl;
				}
			}
			
			segments[ii].caUnits[jj].Buff_Ca_TnCL = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+10];
			segments[ii].caUnits[jj].Buff_Ca_TnCHc = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+11];
			segments[ii].caUnits[jj].Buff_Ca_TnCHm = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+12];
			segments[ii].caUnits[jj].Buff_Ca_CaM = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+13];
			segments[ii].caUnits[jj].Buff_Ca_Myosin_ca = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+14];
			segments[ii].caUnits[jj].Buff_Ca_Myosin_mg = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+15];;
			segments[ii].caUnits[jj].Buff_Ca_SRB = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+16];
			segments[ii].caUnits[jj].Buff_Ca_CSQN = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+17];
	
			segments[ii].caUnits[jj].Ca_i = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+18];
			segments[ii].caUnits[jj].Ca_sr = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+19];
			segments[ii].caUnits[jj].Buff_Ca_EGTA_cyt = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+20];
			
			segments[ii].caUnits[jj].Ca_srs = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+21];
			segments[ii].caUnits[jj].Buff_Ca_SLLsrs = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+22];
			segments[ii].caUnits[jj].Buff_Ca_SLHsrs = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+23];
			
			segments[ii].caUnits[jj].RyR_Integr = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+24];
			for(int abcd=0; abcd<8; abcd++)
			{
				segments[ii].caUnits[jj].ICaLState_NP[abcd] = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+25+abcd];
			}
			
			segments[ii].caUnits[jj].Na_i = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+33];
			segments[ii].caUnits[jj].Na_srs = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+34];
			segments[ii].caUnits[jj].Buff_Na_junc = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+jj*numVarsPerCaDomain+startat+35];
		}
		
		// Other stuff
		segments[ii].K_i = x0data[ii*numVarsPerSegment+2*numVarsPerMembr+settings->N_CaDomains*numVarsPerCaDomain+startat+2];
	}
}
